from datetime import timedelta

from django.core.exceptions import ValidationError
from django.test import override_settings
from django.utils import timezone

from scheduler import settings
from scheduler.models import TaskType, Task
from scheduler.redis_models import JobModel
from scheduler.tests.test_task_types.test_task_model import BaseTestCases
from scheduler.tests.testtools import task_factory, _get_task_scheduled_job_from_registry
from scheduler.types import SchedulerConfiguration


class TestRepeatableTask(BaseTestCases.TestSchedulableTask):
    task_type = TaskType.REPEATABLE
    queue_name = settings.get_queue_names()[0]

    def test_create_task_error(self):
        scheduled_time = timezone.now()

        Task.objects.create(
            name="konichiva_every_2s",
            callable="chat.task_scheduler.konichiva_func",
            task_type="REPEATABLE",
            interval=2,
            interval_unit="seconds",
            queue="default",
            enabled=True,
            scheduled_time=scheduled_time,
        )

    def test_unschedulable_old_job(self):
        job = task_factory(self.task_type, scheduled_time=timezone.now() - timedelta(hours=1), repeat=0)
        self.assertFalse(job.is_scheduled())

    def test_schedulable_old_job_repeat_none(self):
        # If repeat is None, the job should be scheduled
        job = task_factory(self.task_type, scheduled_time=timezone.now() - timedelta(hours=1), repeat=None)
        self.assertTrue(job.is_scheduled())

    def test_clean(self):
        job = task_factory(self.task_type)
        job.queue = self.queue_name
        job.callable = "scheduler.tests.jobs.test_job"
        job.interval = 1
        job.success_ttl = -1
        self.assertIsNone(job.clean())

    def test_clean_seconds(self):
        job = task_factory(self.task_type)
        job.queue = self.queue_name
        job.callable = "scheduler.tests.jobs.test_job"
        job.interval = 60
        job.success_ttl = -1
        job.interval_unit = "seconds"
        self.assertIsNone(job.clean())

    @override_settings(SCHEDULER_CONFIG=SchedulerConfiguration(SCHEDULER_INTERVAL=10))
    def test_clean_too_frequent(self):
        job = task_factory(self.task_type)
        job.queue = self.queue_name
        job.callable = "scheduler.tests.jobs.test_job"
        job.interval = 2  # Smaller than 10
        job.success_ttl = -1
        job.interval_unit = "seconds"
        with self.assertRaises(ValidationError):
            job.clean_interval_unit()

    @override_settings(SCHEDULER_CONFIG=SchedulerConfiguration(SCHEDULER_INTERVAL=10))
    def test_clean_not_multiple(self):
        job = task_factory(self.task_type)
        job.queue = self.queue_name
        job.callable = "scheduler.tests.jobs.test_job"
        job.interval = 121
        job.interval_unit = "seconds"
        with self.assertRaises(ValidationError):
            job.clean_interval_unit()

    def test_clean_short_result_ttl(self):
        task = task_factory(self.task_type)
        task.queue = self.queue_name
        task.callable = "scheduler.tests.jobs.test_job"
        task.interval = 1
        task.repeat = 1
        task.result_ttl = 3599
        task.interval_unit = "hours"
        task.repeat = 42
        with self.assertRaises(ValidationError):
            task.clean_result_ttl()

    def test_clean_indefinite_result_ttl(self):
        task = task_factory(self.task_type)
        task.queue = self.queue_name
        task.callable = "scheduler.tests.jobs.test_job"
        task.interval = 1
        task.result_ttl = -1
        task.interval_unit = "hours"
        task.clean_result_ttl()

    def test_clean_undefined_result_ttl(self):
        task = task_factory(self.task_type)
        task.queue = self.queue_name
        task.callable = "scheduler.tests.jobs.test_job"
        task.interval = 1
        task.interval_unit = "hours"
        task.clean_result_ttl()

    def test_interval_seconds_weeks(self):
        task = task_factory(self.task_type, interval=2, interval_unit="weeks")
        self.assertEqual(1209600.0, task.interval_seconds())

    def test_interval_seconds_days(self):
        task = task_factory(self.task_type, interval=2, interval_unit="days")
        self.assertEqual(172800.0, task.interval_seconds())

    def test_interval_seconds_hours(self):
        job = task_factory(self.task_type, interval=2, interval_unit="hours")
        self.assertEqual(7200.0, job.interval_seconds())

    def test_interval_seconds_minutes(self):
        job = task_factory(self.task_type, interval=15, interval_unit="minutes")
        self.assertEqual(900.0, job.interval_seconds())

    def test_interval_seconds_seconds(self):
        job = task_factory(self.task_type, interval=15, interval_unit="seconds")
        self.assertEqual(15.0, job.interval_seconds())

    def test_result_interval(self):
        job = task_factory(self.task_type)
        entry = _get_task_scheduled_job_from_registry(job)
        self.assertEqual(entry.meta["interval"], 3600)

    def test_repeat(self):
        job = task_factory(self.task_type, repeat=10)
        entry = _get_task_scheduled_job_from_registry(job)
        self.assertEqual(entry.meta["repeat"], 10)

    def test_repeat_old_job_exhausted(self):
        base_time = timezone.now()
        job = task_factory(self.task_type, scheduled_time=base_time - timedelta(hours=10), repeat=10)
        self.assertEqual(job.is_scheduled(), False)

    def test_repeat_old_job_last_iter(self):
        base_time = timezone.now()
        job = task_factory(self.task_type, scheduled_time=base_time - timedelta(hours=9, minutes=30), repeat=10)
        self.assertEqual(job.repeat, 0)
        self.assertEqual(job.is_scheduled(), True)

    def test_repeat_old_job_remaining(self):
        base_time = timezone.now()
        job = task_factory(self.task_type, scheduled_time=base_time - timedelta(minutes=30), repeat=5)
        self.assertEqual(job.repeat, 4)
        self.assertEqual(job.scheduled_time, base_time + timedelta(minutes=30))
        self.assertEqual(job.is_scheduled(), True)

    def test_repeat_none_interval_2_min(self):
        base_time = timezone.now()
        task = task_factory(self.task_type, scheduled_time=base_time - timedelta(minutes=29), repeat=None)
        task.interval = 120
        task.interval_unit = "seconds"
        task.save()
        self.assertTrue(task.scheduled_time > base_time)
        self.assertTrue(task.is_scheduled())

    def test_check_rescheduled_after_execution(self):
        task = task_factory(self.task_type, scheduled_time=timezone.now() + timedelta(seconds=1), repeat=10)
        queue = task.rqueue
        first_run_id = task.job_name
        entry = JobModel.get(first_run_id, connection=queue.connection)
        queue.run_sync(entry)
        task.refresh_from_db()
        self.assertEqual(task.failed_runs, 0)
        self.assertIsNone(task.last_failed_run)
        self.assertEqual(task.successful_runs, 1)
        self.assertIsNotNone(task.last_successful_run)
        self.assertTrue(task.is_scheduled())
        self.assertNotEqual(task.job_name, first_run_id)

    def test_check_rescheduled_after_execution_failed_job(self):
        task = task_factory(
            self.task_type,
            callable_name="scheduler.tests.jobs.failing_job",
            scheduled_time=timezone.now() + timedelta(seconds=1),
            repeat=10,
        )
        queue = task.rqueue
        first_run_id = task.job_name
        entry = JobModel.get(first_run_id, connection=queue.connection)
        queue.run_sync(entry)
        task.refresh_from_db()
        self.assertEqual(task.failed_runs, 1)
        self.assertIsNotNone(task.last_failed_run)
        self.assertEqual(task.successful_runs, 0)
        self.assertIsNone(task.last_successful_run)
        self.assertTrue(task.is_scheduled())
        self.assertNotEqual(task.job_name, first_run_id)

    def test_check_not_rescheduled_after_last_repeat(self):
        task = task_factory(
            self.task_type,
            scheduled_time=timezone.now() + timedelta(seconds=1),
            repeat=1,
        )
        queue = task.rqueue
        first_run_id = task.job_name
        entry = JobModel.get(first_run_id, connection=queue.connection)
        queue.run_sync(entry)
        task.refresh_from_db()
        self.assertEqual(task.failed_runs, 0)
        self.assertIsNone(task.last_failed_run)
        self.assertEqual(task.successful_runs, 1)
        self.assertIsNotNone(task.last_successful_run)
        self.assertNotEqual(task.job_name, first_run_id)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class SchedulerConfig(AppConfig):
    default_auto_field = "django.db.models.AutoField"
    name = "scheduler"
    verbose_name = _("Tasks Scheduler")

    def ready(self):
        pass



__all__ = [
    "Task",
    "TaskType",
    "TaskArg",
    "TaskKwarg",
    "get_scheduled_task",
    "run_task",
    "get_next_cron_time",
]

from .args import TaskArg, TaskKwarg
from .task import TaskType, Task, get_scheduled_task, run_task, get_next_cron_time



from datetime import timedelta

from django.core.exceptions import ImproperlyConfigured
from django.test import override_settings
from django.utils import timezone

from scheduler.helpers.callback import Callback, CallbackSetupError
from scheduler.models import TaskType, get_scheduled_task
from scheduler.tests.testtools import SchedulerBaseCase, task_factory


class TestInternals(SchedulerBaseCase):
    def test_get_scheduled_job(self):
        task = task_factory(TaskType.ONCE, scheduled_time=timezone.now() - timedelta(hours=1))
        self.assertEqual(task, get_scheduled_task(TaskType.ONCE, task.id))
        with self.assertRaises(ValueError):
            get_scheduled_task(task.task_type, task.id + 1)
        with self.assertRaises(ValueError):
            get_scheduled_task("UNKNOWN_JOBTYPE", task.id)

    def test_task_update(self):
        task = task_factory(TaskType.ONCE)
        task.name = "new_name"
        task.save(update_fields=["name"])

    def test_callback_bad_arguments(self):
        with self.assertRaises(CallbackSetupError) as cm:
            Callback("scheduler.tests.jobs.test_job", "1m")
        self.assertEqual(str(cm.exception), "Callback `timeout` must be a positive int, but received 1m")
        with self.assertRaises(CallbackSetupError) as cm:
            Callback("scheduler.tests.jobs.non_existing_method")
        self.assertEqual(str(cm.exception), "Callback `func` is not callable: scheduler.tests.jobs.non_existing_method")
        with self.assertRaises(CallbackSetupError) as cm:
            Callback("scheduler.tests.non_existing_module.non_existing_method")
        self.assertEqual(
            str(cm.exception),
            "Callback `func` is not callable: scheduler.tests.non_existing_module.non_existing_method",
        )
        with self.assertRaises(CallbackSetupError) as cm:
            Callback("non_existing_method")
        self.assertEqual(str(cm.exception), "Callback `func` is not callable: non_existing_method")
        with self.assertRaises(CallbackSetupError) as cm:
            Callback(1)
        self.assertEqual(str(cm.exception), "Callback `func` must be a string or function, received 1")


class TestConfSettings(SchedulerBaseCase):
    @override_settings(SCHEDULER_CONFIG=[])
    def test_conf_settings__bad_scheduler_config(self):
        from scheduler import settings

        with self.assertRaises(ImproperlyConfigured) as cm:
            settings.conf_settings()

        self.assertEqual(str(cm.exception), "SCHEDULER_CONFIG should be a SchedulerConfiguration or dict")

    @override_settings(SCHEDULER_QUEUES=[])
    def test_conf_settings__bad_scheduler_queues_config(self):
        from scheduler import settings

        with self.assertRaises(ImproperlyConfigured) as cm:
            settings.conf_settings()

        self.assertEqual(str(cm.exception), "You have to define SCHEDULER_QUEUES in settings.py as dict")

    @override_settings(SCHEDULER_QUEUES={"default": []})
    def test_conf_settings__bad_queue_config(self):
        from scheduler import settings

        with self.assertRaises(ImproperlyConfigured) as cm:
            settings.conf_settings()

        self.assertEqual(str(cm.exception), "Queue default configuration should be a QueueConfiguration or dict")

    @override_settings(SCHEDULER_CONFIG={"UNKNOWN_SETTING": 10})
    def test_conf_settings__unknown_setting(self):
        from scheduler import settings

        with self.assertRaises(ImproperlyConfigured) as cm:
            settings.conf_settings()

        self.assertEqual(str(cm.exception), "Unknown setting UNKNOWN_SETTING in SCHEDULER_CONFIG")



from django.contrib import admin

from scheduler import views
from scheduler.models.ephemeral_models import Queue, Worker


class ImmutableAdmin(admin.ModelAdmin):
    def has_add_permission(self, request):
        return False  # Hide the admin "+ Add" link for Queues

    def has_change_permission(self, request, obj=None):
        return True

    def has_module_permission(self, request):
        """Returns True if the given request has any permission in the given app label.

        Can be overridden by the user in subclasses. In such case, it should return True if the given request has
        permission to view the module on the admin index page and access the module's index page. Overriding it does
        not restrict access to the add, change or delete views. Use `ModelAdmin.has_(add|change|delete)_permission` for
        that.
        """
        return request.user.has_module_perms("django-tasks-scheduler")


@admin.register(Queue)
class QueueAdmin(ImmutableAdmin):
    """Admin View for queues"""

    def changelist_view(self, request, extra_context=None):
        """The 'change list' admin view for this model."""
        return views.stats(request)


@admin.register(Worker)
class WorkerAdmin(ImmutableAdmin):
    """Admin View for workers"""

    def changelist_view(self, request, extra_context=None):
        """The 'change list' admin view for this model."""
        return views.workers_list(request)



__all__ = [
    "WorkerCommandsChannelListener",
    "StopJobCommand",
    "ShutdownCommand",
    "KillWorkerCommand",
    "WorkerCommandError",
    "send_command",
]

from .kill_worker import KillWorkerCommand
from .shutdown import ShutdownCommand
from .stop_job import StopJobCommand
from .worker_commands import WorkerCommandsChannelListener, WorkerCommandError, send_command



from scheduler.types import ConnectionType
from scheduler.redis_models import WorkerModel
from scheduler.settings import logger
from scheduler.worker.commands.worker_commands import WorkerCommand


class SuspendWorkCommand(WorkerCommand):
    """Suspend worker command"""

    command_name = "suspend"

    def process_command(self, connection: ConnectionType) -> None:
        logger.debug(f"Received command to suspend worker {self.worker_name}")
        worker_model = WorkerModel.get(self.worker_name, connection)
        if worker_model is None:
            logger.warning(f"Worker {self.worker_name} not found")
        if worker_model.is_suspended:
            logger.warning(f"Worker {self.worker_name} already suspended")
            return
        worker_model.set_field("is_suspended", True, connection=connection)
        logger.info(f"Worker {self.worker_name} suspended")


class ResumeWorkCommand(WorkerCommand):
    """Resume worker command"""

    command_name = "resume"

    def process_command(self, connection: ConnectionType) -> None:
        logger.debug(f"Received command to resume worker {self.worker_name}")
        worker_model = WorkerModel.get(self.worker_name, connection)
        if worker_model is None:
            logger.warning(f"Worker {self.worker_name} not found")
        if not worker_model.is_suspended:
            logger.warning(f"Worker {self.worker_name} not suspended and therefore can't be resumed")
            return
        worker_model.set_field("is_suspended", False, connection=connection)
        logger.info(f"Worker {self.worker_name} resumed")



import errno
import os
import signal
from typing import Optional

from scheduler.types import ConnectionType
from scheduler.redis_models import WorkerModel
from scheduler.settings import logger
from scheduler.worker.commands.worker_commands import WorkerCommand


class KillWorkerCommand(WorkerCommand):
    """kill-worker command"""

    command_name = "kill-worker"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.worker_pid: Optional[int] = None

    def process_command(self, connection: ConnectionType) -> None:
        from scheduler.worker import Worker

        logger.info("Received kill-worker command.")
        worker_model = WorkerModel.get(self.worker_name, connection)
        self.worker_pid = worker_model.pid
        if self.worker_pid is None:
            raise ValueError("Worker PID is not set")
        logger.info(f"Killing worker main process {self.worker_pid}...")
        try:
            Worker.from_model(worker_model).request_stop(signal.SIGTERM, None)
            os.killpg(os.getpgid(self.worker_pid), signal.SIGTERM)
            logger.info(f"Killed worker main process pid {self.worker_pid}")
        except OSError as e:
            if e.errno == errno.ESRCH:
                logger.debug(
                    f"Worker main process for {self.worker_name}:{self.worker_pid} already dead"
                )  # "No such process" is fine with us
            else:
                raise



from django.urls import reverse

from scheduler.helpers.queues import get_queue
from scheduler.worker import create_worker
from scheduler.redis_models import JobStatus, JobModel
from scheduler.tests.jobs import failing_job, long_job, test_job
from scheduler.tests.testtools import assert_message_in_response
from .base import BaseTestCase
from ..test_task_types.test_task_model import assert_response_has_msg


class SingleJobActionViewsTest(BaseTestCase):
    def test_single_job_action_unknown_job(self):
        res = self.client.get(reverse("job_detail_action", args=["unknown", "cancel"]), follow=True)
        self.assertEqual(200, res.status_code)
        assert_response_has_msg(res, "Job unknown does not exist, maybe its TTL has passed")

    def test_single_job_action_unknown_action(self):
        queue = get_queue("default")
        job = queue.create_and_enqueue_job(failing_job)
        worker = create_worker("default", burst=True)
        worker.work()
        job = JobModel.get(job.name, connection=queue.connection)
        self.assertTrue(job.is_failed)
        res = self.client.get(reverse("job_detail_action", args=[job.name, "unknown"]), follow=True)
        self.assertEqual(400, res.status_code)

    def test_single_job_action_delete_job(self):
        queue = get_queue("default")
        job = queue.create_and_enqueue_job(test_job, job_info_ttl=0)
        res = self.client.get(reverse("job_detail_action", args=[job.name, "delete"]), follow=True)
        self.assertEqual(200, res.status_code)
        self.client.post(reverse("job_detail_action", args=[job.name, "delete"]), {"post": "yes"}, follow=True)
        self.assertFalse(JobModel.exists(job.name, connection=queue.connection))
        self.assertNotIn(job.name, queue.queued_job_registry.all())

    def test_single_job_action_cancel_job(self):
        queue = get_queue("django_tasks_scheduler_test")
        job = queue.create_and_enqueue_job(long_job)
        self.assertTrue(job.is_queued)
        job = JobModel.get(job.name, connection=queue.connection)
        res = self.client.get(reverse("job_detail_action", args=[job.name, "cancel"]), follow=True)
        self.assertEqual(200, res.status_code)
        res = self.client.post(reverse("job_detail_action", args=[job.name, "cancel"]), {"post": "yes"}, follow=True)
        self.assertEqual(200, res.status_code)
        job = JobModel.get(job.name, connection=queue.connection)
        self.assertTrue(job.is_canceled)
        self.assertNotIn(job.name, queue.queued_job_registry.all())

    def test_single_job_action_cancel_job_that_is_already_cancelled(self):
        queue = get_queue("django_tasks_scheduler_test")
        job = queue.create_and_enqueue_job(long_job)
        res = self.client.post(reverse("job_detail_action", args=[job.name, "cancel"]), {"post": "yes"}, follow=True)
        self.assertEqual(200, res.status_code)
        tmp = JobModel.get(job.name, connection=queue.connection)
        self.assertTrue(tmp.is_canceled)
        self.assertNotIn(job.name, queue.queued_job_registry.all())
        res = self.client.post(reverse("job_detail_action", args=[job.name, "cancel"]), {"post": "yes"}, follow=True)
        self.assertEqual(200, res.status_code)
        assert_message_in_response(res, f"Could not perform action: Cannot cancel already canceled job: {job.name}")

    def test_single_job_action_enqueue_job(self):
        queue = get_queue("django_tasks_scheduler_test")
        job_list = []
        # enqueue some jobs that depends on other
        for _ in range(0, 3):
            job = queue.create_and_enqueue_job(test_job)
            job_list.append(job)

        self.assertEqual(job_list[-1].get_status(connection=queue.connection), JobStatus.QUEUED)
        self.assertIsNotNone(job_list[-1].enqueued_at)

        # Try to force enqueue last job should do nothing
        res = self.client.get(reverse("job_detail_action", args=[job_list[-1].name, "enqueue"]), follow=True)
        self.assertEqual(200, res.status_code)
        res = self.client.post(reverse("job_detail_action", args=[job_list[-1].name, "enqueue"]), follow=True)

        self.assertEqual(200, res.status_code)
        tmp = JobModel.get(job_list[-1].name, connection=queue.connection)
        self.assertEqual(tmp.get_status(connection=queue.connection), JobStatus.QUEUED)
        self.assertIsNotNone(tmp.enqueued_at)

    def test_single_job_action_enqueue_job_sync_queue(self):
        queue = get_queue("scheduler_scheduler_active_test")
        job_list = []
        # enqueue some jobs that depends on other
        for _ in range(0, 3):
            job = queue.create_and_enqueue_job(test_job)
            job_list.append(job)

        self.assertEqual(job_list[-1].status, JobStatus.FINISHED)
        self.assertIsNotNone(job_list[-1].enqueued_at)

        # Try to force enqueue last job should do nothing
        res = self.client.get(reverse("job_detail_action", args=[job_list[-1].name, "enqueue"]), follow=True)
        self.assertEqual(200, res.status_code)
        res = self.client.post(reverse("job_detail_action", args=[job_list[-1].name, "enqueue"]), follow=True)

        self.assertEqual(200, res.status_code)
        tmp = JobModel.get(job_list[-1].name, connection=queue.connection)
        self.assertEqual(tmp.get_status(connection=queue.connection), JobStatus.FINISHED)
        self.assertIsNotNone(tmp.enqueued_at)
        self.assertGreater(tmp.enqueued_at, job_list[-1].enqueued_at)



import dataclasses
from math import ceil
from typing import Tuple, List, Dict, Union, Any, Optional

from django.contrib import admin
from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponse, HttpRequest, HttpResponseNotFound, JsonResponse
from django.shortcuts import render
from django.views.decorators.cache import never_cache

from scheduler import settings
from scheduler.helpers.queues import Queue, get_all_workers
from scheduler.redis_models import JobModel, JobNamesRegistry, WorkerModel
from scheduler.settings import get_queue_names, logger
from scheduler.types import ConnectionErrorTypes
from scheduler.views.helpers import get_queue


def _get_registry_job_list(queue: Queue, registry: JobNamesRegistry, page: int) -> Tuple[List[JobModel], int, range]:
    items_per_page = settings.SCHEDULER_CONFIG.EXECUTIONS_IN_PAGE
    num_jobs = registry.count(queue.connection)
    job_list = list()

    if num_jobs == 0:
        return job_list, num_jobs, range(1, 1)

    last_page = int(ceil(num_jobs / items_per_page))
    page_range = range(1, last_page + 1)
    offset = items_per_page * (page - 1)
    job_names = registry.all(offset, offset + items_per_page - 1)
    job_list = JobModel.get_many(job_names, connection=queue.connection)
    remove_job_names = [job_name for i, job_name in enumerate(job_names) if job_list[i] is None]
    valid_jobs = [job for job in job_list if job is not None]
    if registry is not queue:
        for job_name in remove_job_names:
            registry.delete(queue.connection, job_name)

    return valid_jobs, num_jobs, page_range


@never_cache
@staff_member_required
def list_registry_jobs(request: HttpRequest, queue_name: str, registry_name: str) -> HttpResponse:
    queue = get_queue(queue_name)
    registry = queue.get_registry(registry_name)
    if registry is None:
        return HttpResponseNotFound()
    title = registry_name.capitalize()
    page = int(request.GET.get("page", 1))
    job_list, num_jobs, page_range = _get_registry_job_list(queue, registry, page)

    context_data = {
        **admin.site.each_context(request),
        "queue": queue,
        "registry_name": registry_name,
        "registry": registry,
        "jobs": job_list,
        "num_jobs": num_jobs,
        "page": page,
        "page_range": page_range,
        "job_status": title,
    }
    return render(request, "admin/scheduler/jobs.html", context_data)


@never_cache
@staff_member_required
def queue_workers(request: HttpRequest, queue_name: str) -> HttpResponse:
    queue = get_queue(queue_name)
    queue.clean_registries()

    all_workers = WorkerModel.all(queue.connection)
    worker_list = [worker for worker in all_workers if queue.name in worker.queue_names]

    context_data = {
        **admin.site.each_context(request),
        "queue": queue,
        "workers": worker_list,
    }
    return render(request, "admin/scheduler/queue_workers.html", context_data)


def stats_json(request: HttpRequest) -> Union[JsonResponse, HttpResponseNotFound]:
    auth_token = request.headers.get("Authorization")
    token_validation_func = settings.SCHEDULER_CONFIG.TOKEN_VALIDATION_METHOD
    if request.user.is_staff or (token_validation_func and auth_token and token_validation_func(auth_token)):
        return JsonResponse(get_statistics())

    return HttpResponseNotFound()


@never_cache
@staff_member_required
def stats(request: HttpRequest) -> HttpResponse:
    context_data = {**admin.site.each_context(request), **get_statistics(run_maintenance_tasks=True)}
    return render(request, "admin/scheduler/stats.html", context_data)


@dataclasses.dataclass
class QueueData:
    name: str
    queued_jobs: int
    oldest_job_timestamp: Optional[str]
    scheduler_pid: Optional[int]
    workers: int
    finished_jobs: int
    started_jobs: int
    failed_jobs: int
    scheduled_jobs: int
    canceled_jobs: int


def get_statistics(run_maintenance_tasks: bool = False) -> Dict[str, List[Dict[str, Any]]]:
    queue_names = get_queue_names()
    queues: List[QueueData] = []
    queue_workers_count: Dict[str, int] = {queue_name: 0 for queue_name in queue_names}
    workers = get_all_workers()
    for worker in workers:
        for queue_name in worker.queue_names:
            if queue_name not in queue_workers_count:
                logger.warning(f"Worker {worker.name} is working on queue {queue_name} which is not defined")
                queue_workers_count[queue_name] = 0
            queue_workers_count[queue_name] += 1
    for queue_name in queue_names:
        try:
            queue = get_queue(queue_name)
            connection_kwargs = queue.connection.connection_pool.connection_kwargs

            if run_maintenance_tasks:
                queue.clean_registries()

            # Raw access to the first item from left of the broker list.
            # This might not be accurate since new job can be added from the left
            # with `at_front` parameters.
            # Ideally rq should supports Queue.oldest_job

            last_job_name = queue.first_queued_job_name()
            last_job = JobModel.get(last_job_name, connection=queue.connection) if last_job_name else None
            if last_job and last_job.enqueued_at:
                oldest_job_timestamp = last_job.enqueued_at.isoformat()
            else:
                oldest_job_timestamp = None

            # parse_class and connection_pool are not needed and not JSON serializable
            connection_kwargs.pop("parser_class", None)
            connection_kwargs.pop("connection_pool", None)

            queue_data = QueueData(
                name=queue.name,
                queued_jobs=len(queue.queued_job_registry),
                oldest_job_timestamp=oldest_job_timestamp,
                scheduler_pid=queue.scheduler_pid,
                workers=queue_workers_count[queue.name],
                finished_jobs=len(queue.finished_job_registry),
                started_jobs=len(queue.active_job_registry),
                failed_jobs=len(queue.failed_job_registry),
                scheduled_jobs=len(queue.scheduled_job_registry),
                canceled_jobs=len(queue.canceled_job_registry),
            )
            queues.append(queue_data)
        except ConnectionErrorTypes as e:
            logger.error(f"Could not connect for queue {queue_name}: {e}")
            continue

    return {"queues": [dataclasses.asdict(q) for q in queues]}



import json
import os
import tempfile
from unittest import mock

import yaml
from django.core.management import call_command
from django.test import TestCase

from scheduler.models import Task
from scheduler.tests import test_settings  # noqa
from scheduler.tests.testtools import task_factory
from scheduler.models import TaskType


class ImportTest(TestCase):
    def setUp(self) -> None:
        self.tmpfile = tempfile.NamedTemporaryFile(mode="w")

    def tearDown(self) -> None:
        os.remove(self.tmpfile.name)

    def test_import__should_schedule_job(self):
        jobs = list()
        jobs.append(task_factory(TaskType.ONCE, enabled=True, instance_only=True))
        jobs.append(task_factory(TaskType.REPEATABLE, enabled=True, instance_only=True))
        res = json.dumps([j.to_dict() for j in jobs])
        self.tmpfile.write(res)
        self.tmpfile.flush()
        # act
        call_command("import", filename=self.tmpfile.name)
        # assert
        self.assertEqual(1, Task.objects.filter(task_type=TaskType.ONCE).count())
        self.assertEqual(1, Task.objects.filter(task_type=TaskType.REPEATABLE).count())
        db_job = Task.objects.filter(task_type=TaskType.ONCE).first()
        attrs = ["name", "queue", "callable", "enabled", "timeout"]
        for attr in attrs:
            self.assertEqual(getattr(jobs[0], attr), getattr(db_job, attr))

    def test_import__should_schedule_job_yaml(self):
        tasks = list()
        tasks.append(task_factory(TaskType.ONCE, enabled=True, instance_only=True))
        tasks.append(task_factory(TaskType.REPEATABLE, enabled=True, instance_only=True))
        res = yaml.dump([j.to_dict() for j in tasks], default_flow_style=False)
        self.tmpfile.write(res)
        self.tmpfile.flush()
        # act
        call_command("import", filename=self.tmpfile.name, format="yaml")
        # assert
        self.assertEqual(1, Task.objects.filter(task_type=TaskType.ONCE).count())
        self.assertEqual(1, Task.objects.filter(task_type=TaskType.REPEATABLE).count())
        db_job = Task.objects.filter(task_type=TaskType.ONCE).first()
        attrs = ["name", "queue", "callable", "enabled", "timeout"]
        for attr in attrs:
            self.assertEqual(getattr(tasks[0], attr), getattr(db_job, attr))

    def test_import__should_schedule_job_yaml_without_yaml_lib(self):
        jobs = list()
        jobs.append(task_factory(TaskType.ONCE, enabled=True, instance_only=True))
        jobs.append(task_factory(TaskType.REPEATABLE, enabled=True, instance_only=True))
        res = yaml.dump([j.to_dict() for j in jobs], default_flow_style=False)
        self.tmpfile.write(res)
        self.tmpfile.flush()
        # act
        with mock.patch.dict("sys.modules", {"yaml": None}):
            with self.assertRaises(SystemExit) as cm:
                call_command("import", filename=self.tmpfile.name, format="yaml")
            self.assertEqual(cm.exception.code, 1)

    def test_import__should_schedule_job_reset(self):
        jobs = list()
        task_factory(TaskType.ONCE, enabled=True)
        task_factory(TaskType.ONCE, enabled=True)
        jobs.append(task_factory(TaskType.ONCE, enabled=True))
        jobs.append(task_factory(TaskType.REPEATABLE, enabled=True, instance_only=True))
        res = json.dumps([j.to_dict() for j in jobs])
        self.tmpfile.write(res)
        self.tmpfile.flush()
        # act
        call_command(
            "import",
            filename=self.tmpfile.name,
            reset=True,
        )
        # assert
        self.assertEqual(1, Task.objects.filter(task_type=TaskType.ONCE).count())
        db_job = Task.objects.filter(task_type=TaskType.ONCE).first()
        attrs = ["name", "queue", "callable", "enabled", "timeout"]
        for attr in attrs:
            self.assertEqual(getattr(jobs[0], attr), getattr(db_job, attr))
        self.assertEqual(1, Task.objects.filter(task_type=TaskType.REPEATABLE).count())
        db_job = Task.objects.filter(task_type=TaskType.REPEATABLE).first()
        attrs = ["name", "queue", "callable", "enabled", "timeout"]
        for attr in attrs:
            self.assertEqual(getattr(jobs[1], attr), getattr(db_job, attr))

    def test_import__should_schedule_job_update_existing(self):
        tasks = list()
        tasks.append(task_factory(TaskType.ONCE, enabled=True))
        tasks.append(task_factory(TaskType.ONCE, enabled=True))
        res = json.dumps([j.to_dict() for j in tasks])
        self.tmpfile.write(res)
        self.tmpfile.flush()
        # act
        call_command(
            "import",
            filename=self.tmpfile.name,
            update=True,
        )
        # assert
        self.assertEqual(2, Task.objects.filter(task_type=TaskType.ONCE).count())
        db_job = Task.objects.filter(task_type=TaskType.ONCE).get(name=tasks[0].name)
        attrs = ["name", "queue", "callable", "enabled", "timeout"]
        for attr in attrs:
            self.assertEqual(getattr(tasks[0], attr), getattr(db_job, attr))

    def test_import__should_schedule_job_without_update_existing(self):
        tasks = list()
        tasks.append(task_factory(TaskType.ONCE, enabled=True))
        tasks.append(task_factory(TaskType.ONCE, enabled=True))
        res = json.dumps([j.to_dict() for j in tasks])
        self.tmpfile.write(res)
        self.tmpfile.flush()
        # act
        call_command(
            "import",
            filename=self.tmpfile.name,
        )
        # assert
        self.assertEqual(2, Task.objects.filter(task_type=TaskType.ONCE).count())
        db_job = Task.objects.get(name=tasks[0].name)
        attrs = ["id", "name", "queue", "callable", "enabled", "timeout"]
        for attr in attrs:
            self.assertEqual(getattr(tasks[0], attr), getattr(db_job, attr))



import base64
import dataclasses
import inspect
import numbers
import pickle
from datetime import datetime
from enum import Enum
from typing import ClassVar, Dict, Optional, List, Callable, Any, Union, Tuple

from scheduler.helpers import utils
from scheduler.helpers.callback import Callback
from scheduler.redis_models.base import HashModel, as_str
from scheduler.settings import SCHEDULER_CONFIG, logger
from scheduler.types import ConnectionType, Self, FunctionReferenceType
from .registry.base_registry import JobNamesRegistry
from ..helpers.utils import current_timestamp


class TimeoutFormatError(Exception):
    pass


class JobStatus(str, Enum):
    """The Status of Job within its lifecycle at any given time."""

    QUEUED = "queued"
    FINISHED = "finished"
    FAILED = "failed"
    STARTED = "started"
    SCHEDULED = "scheduled"
    STOPPED = "stopped"
    CANCELED = "canceled"


@dataclasses.dataclass(slots=True, kw_only=True)
class JobModel(HashModel):
    _list_key: ClassVar[str] = ":jobs:ALL:"
    _children_key_template: ClassVar[str] = ":{}:jobs:"
    _element_key_template: ClassVar[str] = ":jobs:{}"
    _non_serializable_fields = {"args", "kwargs"}

    args: List[Any]
    kwargs: Dict[str, str]

    queue_name: str
    description: str
    func_name: str

    timeout: int = SCHEDULER_CONFIG.DEFAULT_JOB_TIMEOUT
    success_ttl: int = SCHEDULER_CONFIG.DEFAULT_SUCCESS_TTL
    job_info_ttl: int = SCHEDULER_CONFIG.DEFAULT_JOB_TTL
    status: JobStatus
    created_at: datetime
    meta: Dict[str, str]
    at_front: bool = False
    last_heartbeat: Optional[datetime] = None
    worker_name: Optional[str] = None
    started_at: Optional[datetime] = None
    enqueued_at: Optional[datetime] = None
    ended_at: Optional[datetime] = None
    success_callback_name: Optional[str] = None
    success_callback_timeout: int = SCHEDULER_CONFIG.CALLBACK_TIMEOUT
    failure_callback_name: Optional[str] = None
    failure_callback_timeout: int = SCHEDULER_CONFIG.CALLBACK_TIMEOUT
    stopped_callback_name: Optional[str] = None
    stopped_callback_timeout: int = SCHEDULER_CONFIG.CALLBACK_TIMEOUT
    task_type: Optional[str] = None
    scheduled_task_id: Optional[int] = None

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):  # noqa
        return isinstance(other, self.__class__) and self.name == other.name

    def __str__(self):
        return f"{self.name}: {self.description}"

    def get_status(self, connection: ConnectionType) -> JobStatus:
        return self.get_field("status", connection=connection)

    @property
    def is_queued(self) -> bool:
        return self.status == JobStatus.QUEUED

    @property
    def is_canceled(self) -> bool:
        return self.status == JobStatus.CANCELED

    @property
    def is_failed(self) -> bool:
        return self.status == JobStatus.FAILED

    @property
    def func(self) -> Callable[[Any], Any]:
        return utils.callable_func(self.func_name)

    @property
    def is_scheduled_task(self) -> bool:
        return self.scheduled_task_id is not None

    def expire(self, ttl: int, connection: ConnectionType) -> None:
        """Expire the Job Model if ttl >= 0"""
        if ttl == 0:
            self.delete(connection=connection)
        elif ttl > 0:
            connection.expire(self._key, ttl)

    def persist(self, connection: ConnectionType) -> None:
        connection.persist(self._key)

    def prepare_for_execution(self, worker_name: str, registry: JobNamesRegistry, connection: ConnectionType) -> None:
        """Prepares the job for execution, setting the worker name,
        heartbeat information, status and other metadata before execution begins.
        :param worker_name: The name of the worker
        :param registry: The registry to add the job to
        :param current_pid: The current process id
        :param connection: The connection to the broker
        """
        self.worker_name = worker_name
        self.last_heartbeat = utils.utcnow()
        self.started_at = self.last_heartbeat
        self.status = JobStatus.STARTED
        registry.add(connection, self.name, self.last_heartbeat.timestamp())
        self.save(connection=connection)

    def after_execution(
        self,
        job_info_ttl: int,
        status: JobStatus,
        connection: ConnectionType,
        prev_registry: Optional[JobNamesRegistry] = None,
        new_registry: Optional[JobNamesRegistry] = None,
    ) -> None:
        """After the job is executed, update the status, heartbeat, and other metadata."""
        self.status = status
        self.ended_at = utils.utcnow()
        self.last_heartbeat = self.ended_at
        if prev_registry is not None:
            prev_registry.delete(connection, self.name)
        if new_registry is not None and job_info_ttl != 0:
            new_registry.add(connection, self.name, current_timestamp() + job_info_ttl)
        self.save(connection=connection)

    @property
    def failure_callback(self) -> Optional[Callback]:
        if self.failure_callback_name is None:
            return None
        logger.debug(f"Running failure callbacks for {self.name}")
        return Callback(self.failure_callback_name, self.failure_callback_timeout)

    @property
    def success_callback(self) -> Optional[Callable[..., Any]]:
        if self.success_callback_name is None:
            return None
        logger.debug(f"Running success callbacks for {self.name}")
        return Callback(self.success_callback_name, self.success_callback_timeout)

    @property
    def stopped_callback(self) -> Optional[Callable[..., Any]]:
        if self.stopped_callback_name is None:
            return None
        logger.debug(f"Running stopped callbacks for {self.name}")
        return Callback(self.stopped_callback_name, self.stopped_callback_timeout)

    def get_call_string(self):
        return _get_call_string(self.func_name, self.args, self.kwargs)

    def serialize(self, with_nones: bool = False) -> Dict[str, str]:
        """Serialize the job model to a dictionary."""
        res = super(JobModel, self).serialize(with_nones=with_nones)
        res["args"] = base64.encodebytes(pickle.dumps(self.args)).decode("utf-8")
        res["kwargs"] = base64.encodebytes(pickle.dumps(self.kwargs)).decode("utf-8")
        return res

    @classmethod
    def deserialize(cls, data: Dict[str, Any]) -> Self:
        """Deserialize the job model from a dictionary."""
        res = super(JobModel, cls).deserialize(data)
        res.args = pickle.loads(base64.decodebytes(data.get("args").encode("utf-8")))
        res.kwargs = pickle.loads(base64.decodebytes(data.get("kwargs").encode("utf-8")))
        return res

    @classmethod
    def create(
        cls,
        connection: ConnectionType,
        func: FunctionReferenceType,
        queue_name: str,
        args: Union[List[Any], Optional[Tuple]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        result_ttl: Optional[int] = None,
        job_info_ttl: Optional[int] = None,
        status: Optional[JobStatus] = None,
        description: Optional[str] = None,
        timeout: Optional[int] = None,
        name: Optional[str] = None,
        task_type: Optional[str] = None,
        scheduled_task_id: Optional[int] = None,
        meta: Optional[Dict[str, Any]] = None,
        *,
        on_success: Optional[Callback] = None,
        on_failure: Optional[Callback] = None,
        on_stopped: Optional[Callback] = None,
        at_front: Optional[bool] = None,
    ) -> Self:
        """Creates a new job-model for the given function, arguments, and keyword arguments.
        :returns: A job-model instance.
        """
        args = args or []
        kwargs = kwargs or {}
        timeout = _parse_timeout(timeout) or SCHEDULER_CONFIG.DEFAULT_JOB_TIMEOUT
        if timeout == 0:
            raise ValueError("0 timeout is not allowed. Use -1 for infinite timeout")
        job_info_ttl = _parse_timeout(job_info_ttl if job_info_ttl is not None else SCHEDULER_CONFIG.DEFAULT_JOB_TTL)
        result_ttl = _parse_timeout(result_ttl)
        if not isinstance(args, (tuple, list)):
            raise TypeError(f"{args!r} is not a valid args list")
        if not isinstance(kwargs, dict):
            raise TypeError(f"{kwargs!r} is not a valid kwargs dict")
        if on_success and not isinstance(on_success, Callback):
            raise ValueError("on_success must be a Callback object")
        if on_failure and not isinstance(on_failure, Callback):
            raise ValueError("on_failure must be a Callback object")
        if on_stopped and not isinstance(on_stopped, Callback):
            raise ValueError("on_stopped must be a Callback object")
        if name is not None and JobModel.exists(name, connection=connection):
            raise ValueError(f"Job with name {name} already exists")
        if name is None:
            date_str = utils.utcnow().strftime("%Y%m%d%H%M%S%f")
            name = f"{queue_name}:{scheduled_task_id or ''}:{date_str}"

        if inspect.ismethod(func):
            _func_name = func.__name__

        elif inspect.isfunction(func) or inspect.isbuiltin(func):
            _func_name = f"{func.__module__}.{func.__qualname__}"
        elif isinstance(func, str):
            _func_name = as_str(func)
        elif not inspect.isclass(func) and hasattr(func, "__call__"):  # a callable class instance
            _func_name = "__call__"
        else:
            raise TypeError(f"Expected a callable or a string, but got: {func}")
        description = description or _get_call_string(func, args or [], kwargs or {}, max_length=75)
        job_info_ttl = job_info_ttl if job_info_ttl is not None else SCHEDULER_CONFIG.DEFAULT_JOB_TTL
        model = JobModel(
            created_at=utils.utcnow(),
            name=name,
            queue_name=queue_name,
            description=description,
            func_name=_func_name,
            args=args or [],
            kwargs=kwargs or {},
            at_front=at_front,
            task_type=task_type,
            scheduled_task_id=scheduled_task_id,
            success_callback_name=on_success.name if on_success else None,
            success_callback_timeout=on_success.timeout if on_success else None,
            failure_callback_name=on_failure.name if on_failure else None,
            failure_callback_timeout=on_failure.timeout if on_failure else None,
            stopped_callback_name=on_stopped.name if on_stopped else None,
            stopped_callback_timeout=on_stopped.timeout if on_stopped else None,
            success_ttl=result_ttl,
            job_info_ttl=job_info_ttl,
            timeout=timeout,
            status=status,
            last_heartbeat=None,
            meta=meta or {},
            worker_name=None,
            enqueued_at=None,
            started_at=None,
            ended_at=None,
        )
        model.save(connection=connection)
        return model


def _get_call_string(
    func_name: Optional[str], args: Any, kwargs: Dict[Any, Any], max_length: Optional[int] = None
) -> Optional[str]:
    """
    Returns a string representation of the call, formatted as a regular
    Python function invocation statement. If max_length is not None, truncate
    arguments with representation longer than max_length.

    :param func_name: The function name
    :param args: The function arguments
    :param kwargs: The function kwargs
    :param max_length: The max length of the return string
    :return: A string representation of the function call
    """
    if func_name is None:
        return None

    arg_list = [as_str(_truncate_long_string(repr(arg), max_length)) for arg in args]

    list_kwargs = [f"{k}={as_str(_truncate_long_string(repr(v), max_length))}" for k, v in kwargs.items()]
    arg_list += sorted(list_kwargs)
    args = ", ".join(arg_list)

    return f"{func_name}({args})"


def _truncate_long_string(data: str, max_length: Optional[int] = None) -> str:
    """Truncate arguments with representation longer than max_length"""
    if max_length is None:
        return data
    return (data[:max_length] + "...") if len(data) > max_length else data


def _parse_timeout(timeout: Union[int, float, str]) -> int:
    """Transfer all kinds of timeout format to an integer representing seconds"""
    if not isinstance(timeout, numbers.Integral) and timeout is not None:
        try:
            timeout = int(timeout)
        except ValueError:
            digit, unit = timeout[:-1], (timeout[-1:]).lower()
            unit_second = {"d": 86400, "h": 3600, "m": 60, "s": 1}
            try:
                timeout = int(digit) * unit_second[unit]
            except (ValueError, KeyError):
                raise TimeoutFormatError(
                    "Timeout must be an integer or a string representing an integer, or "
                    'a string with format: digits + unit, unit can be "d", "h", "m", "s", '
                    'such as "1h", "23m".'
                )

    return timeout



from django.contrib.auth.models import User
from django.test import TestCase
from django.test.client import Client

from scheduler.helpers.queues import get_queue
from scheduler.tests import test_settings  # noqa


class BaseTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_superuser("user", password="pass")
        self.client = Client()
        self.client.login(username=self.user.username, password="pass")
        get_queue("django_tasks_scheduler_test").connection.flushall()



from typing import List

from django.contrib import admin
from django.contrib.admin.views.decorators import staff_member_required
from django.core.paginator import Paginator
from django.http import HttpResponse, HttpRequest, Http404
from django.shortcuts import render
from django.views.decorators.cache import never_cache

from scheduler.helpers.queues import get_all_workers
from scheduler.redis_models import WorkerModel, JobModel
from scheduler.settings import SCHEDULER_CONFIG
from scheduler.views.helpers import get_queue


def get_worker_executions(worker: WorkerModel) -> List[JobModel]:
    res = list()
    for queue_name in worker.queue_names:
        queue = get_queue(queue_name)
        curr_jobs = queue.get_all_jobs()
        curr_jobs = [j for j in curr_jobs if j.worker_name == worker.name]
        res.extend(curr_jobs)
    return res


@never_cache
@staff_member_required
def worker_details(request: HttpRequest, name: str) -> HttpResponse:
    workers = get_all_workers()
    worker = next((w for w in workers if w.name == name), None)

    if worker is None:
        raise Http404(f"Couldn't find worker with this ID: {name}")

    execution_list = get_worker_executions(worker)
    paginator = Paginator(execution_list, SCHEDULER_CONFIG.EXECUTIONS_IN_PAGE)
    page_number = request.GET.get("p", 1)
    page_obj = paginator.get_page(page_number)
    page_range = paginator.get_elided_page_range(page_obj.number)
    current_job = None
    if worker.current_job_name is not None:
        queue = get_queue(worker.queue_names[0])
        current_job = JobModel.get(worker.current_job_name, connection=queue.connection)
    context_data = {
        **admin.site.each_context(request),
        "worker": worker,
        "queue_names": ", ".join(worker.queue_names),
        "current_job": current_job,
        "executions": page_obj,
        "page_range": page_range,
        "page_var": "p",
    }
    return render(request, "admin/scheduler/worker_details.html", context_data)


@never_cache
@staff_member_required
def workers_list(request: HttpRequest) -> HttpResponse:
    all_workers = get_all_workers()
    worker_list = [worker for worker in all_workers]

    context_data = {
        **admin.site.each_context(request),
        "workers": worker_list,
    }
    return render(request, "admin/scheduler/workers_list.html", context_data)



import sys
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Dict, Optional, List, Tuple, Any, Type

from scheduler.timeouts import BaseDeathPenalty, UnixSignalDeathPenalty

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class Broker(Enum):
    REDIS = "redis"
    FAKEREDIS = "fakeredis"
    VALKEY = "valkey"


def _token_validation(token: str) -> bool:
    return False


@dataclass(slots=True, kw_only=True)
class SchedulerConfiguration:
    """Configuration for django-tasks-scheduler"""

    EXECUTIONS_IN_PAGE: int = 20
    SCHEDULER_INTERVAL: int = 10
    BROKER: Broker = Broker.REDIS
    TOKEN_VALIDATION_METHOD: Callable[[str], bool] = _token_validation
    CALLBACK_TIMEOUT: int = 60  # Callback timeout in seconds (success/failure/stopped)
    # Default values, can be override per task
    DEFAULT_SUCCESS_TTL: int = 10 * 60  # Time To Live (TTL) in seconds to keep successful job results
    DEFAULT_FAILURE_TTL: int = 365 * 24 * 60 * 60  # Time To Live (TTL) in seconds to keep job failure information
    DEFAULT_JOB_TTL: int = 10 * 60  # Time To Live (TTL) in seconds to keep job information
    DEFAULT_JOB_TIMEOUT: int = 5 * 60  # timeout (seconds) for a job
    # General configuration values
    DEFAULT_WORKER_TTL: int = 10 * 60  # Time To Live (TTL) in seconds to keep worker information after last heartbeat
    DEFAULT_MAINTENANCE_TASK_INTERVAL: int = 10 * 60  # The interval to run maintenance tasks in seconds. 10 minutes.
    DEFAULT_JOB_MONITORING_INTERVAL: int = 30  # The interval to monitor jobs in seconds.
    SCHEDULER_FALLBACK_PERIOD_SECS: int = 120  # Period (secs) to wait before requiring to reacquire locks
    DEATH_PENALTY_CLASS: Type["BaseDeathPenalty"] = UnixSignalDeathPenalty


@dataclass(slots=True, frozen=True, kw_only=True)
class QueueConfiguration:
    __CONNECTION_FIELDS__ = {
        "URL",
        "DB",
        "UNIX_SOCKET_PATH",
        "HOST",
        "PORT",
        "PASSWORD",
        "SENTINELS",
        "MASTER_NAME",
        "CONNECTION_KWARGS",
    }
    DB: Optional[int] = None
    # Redis connection parameters, either UNIX_SOCKET_PATH/URL/separate params (HOST, PORT, PASSWORD) should be provided
    UNIX_SOCKET_PATH: Optional[str] = None
    URL: Optional[str] = None
    HOST: Optional[str] = None
    PORT: Optional[int] = None
    USERNAME: Optional[str] = None
    PASSWORD: Optional[str] = None

    ASYNC: Optional[bool] = True

    SENTINELS: Optional[List[Tuple[str, int]]] = None
    SENTINEL_KWARGS: Optional[Dict[str, str]] = None
    MASTER_NAME: Optional[str] = None
    CONNECTION_KWARGS: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        if not any((self.URL, self.UNIX_SOCKET_PATH, self.HOST, self.SENTINELS)):
            raise ValueError(f"At least one of URL, UNIX_SOCKET_PATH, HOST must be provided: {self}")
        if sum((self.URL is not None, self.UNIX_SOCKET_PATH is not None, self.HOST is not None)) > 1:
            raise ValueError(f"Only one of URL, UNIX_SOCKET_PATH, HOST should be provided: {self}")
        if self.HOST is not None and (self.PORT is None or self.DB is None):
            raise ValueError(f"HOST requires PORT and DB: {self}")

    def same_connection_params(self, other: Self) -> bool:
        for field in self.__CONNECTION_FIELDS__:
            if getattr(self, field) != getattr(other, field):
                return False
        return True



import multiprocessing
from datetime import timedelta
from typing import List, Tuple

from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.contrib.messages import get_messages
from django.test import Client, TestCase
from django.utils import timezone

from scheduler import settings
from scheduler.admin.task_admin import job_execution_of
from scheduler.helpers.queues import get_queue
from scheduler.models import Task, TaskType
from scheduler.models import TaskKwarg
from scheduler.redis_models import JobModel
from scheduler.worker import Worker
from scheduler.worker import create_worker

multiprocessing.set_start_method("fork")


def _run_worker_process(worker: Worker, **kwargs):
    worker.work(**kwargs)


def run_worker_in_process(*args, name="test-worker") -> Tuple[multiprocessing.Process, str]:
    worker = create_worker(*args, name=name, fork_job_execution=False)
    process = multiprocessing.Process(target=_run_worker_process, args=(worker,), kwargs=dict(with_scheduler=False))
    process.start()
    return process, name


def assert_message_in_response(response, message):
    messages = [m.message for m in get_messages(response.wsgi_request)]
    assert message in messages, f'expected "{message}" in {messages}'


def sequence_gen():
    n = 1
    while True:
        yield n
        n += 1


seq = sequence_gen()


def task_factory(
    task_type: TaskType, callable_name: str = "scheduler.tests.jobs.test_job", instance_only=False, **kwargs
):
    values = dict(
        name="Scheduled Job %d" % next(seq),
        queue=list(settings._QUEUES.keys())[0],
        callable=callable_name,
        enabled=True,
        timeout=None,
    )
    if task_type == TaskType.ONCE:
        values.update(
            dict(
                result_ttl=None,
                scheduled_time=timezone.now() + timedelta(days=1),
            )
        )
    elif task_type == TaskType.REPEATABLE:
        values.update(
            dict(
                result_ttl=None,
                interval=1,
                interval_unit="hours",
                repeat=None,
                scheduled_time=timezone.now() + timedelta(days=1),
            )
        )
    elif task_type == TaskType.CRON:
        values.update(
            dict(
                cron_string="0 0 * * *",
            )
        )
    values.update(kwargs)
    if instance_only:
        instance = Task(task_type=task_type, **values)
    else:
        instance = Task.objects.create(task_type=task_type, **values)
    return instance


def taskarg_factory(cls, **kwargs):
    content_object = kwargs.pop("content_object", None)
    if content_object is None:
        content_object = task_factory(TaskType.ONCE)
    values = dict(
        arg_type="str",
        val="",
        object_id=content_object.id,
        content_type=ContentType.objects.get_for_model(content_object),
        content_object=content_object,
    )
    if cls == TaskKwarg:
        values["key"] = ("key%d" % next(seq),)
    values.update(kwargs)
    instance = cls.objects.create(**values)
    return instance


def _get_task_scheduled_job_from_registry(django_task: Task) -> JobModel:
    jobs_to_schedule = django_task.rqueue.scheduled_job_registry.all()
    entry = next(i for i in jobs_to_schedule if i == django_task.job_name)
    return JobModel.get(entry, connection=django_task.rqueue.connection)


def _get_executions(task: Task):
    job_names = task.rqueue.get_all_job_names()
    job_list: List[JobModel] = JobModel.get_many(job_names, connection=task.rqueue.connection)
    return list(filter(lambda j: job_execution_of(j, task), job_list))


class SchedulerBaseCase(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        try:
            User.objects.create_superuser("admin", "admin@a.com", "admin")
        except Exception:
            pass
        cls.client = Client()

    def setUp(self) -> None:
        super(SchedulerBaseCase, self).setUp()
        queue = get_queue("default")
        queue.connection.flushall()

    def tearDown(self) -> None:
        super(SchedulerBaseCase, self).tearDown()



from typing import Dict, Optional

from django import template
from django.utils.safestring import mark_safe

from scheduler.helpers.queues import Queue
from scheduler.models import Task, get_scheduled_task
from scheduler.models.task import run_task
from scheduler.redis_models import Result, JobModel
from scheduler.views.helpers import get_queue

register = template.Library()


@register.filter
def show_func_name(job: JobModel) -> str:
    try:
        res = job.func_name
        if job.func == run_task:
            task = get_scheduled_task(*job.args)
            res = task.function_string()
        return mark_safe(res)
    except Exception as e:
        return repr(e)


@register.filter
def get_item(dictionary: Dict, key):
    return dictionary.get(key)


@register.filter
def scheduled_task(job: JobModel) -> Task:
    django_scheduled_task = get_scheduled_task(*job.args)
    return django_scheduled_task.get_absolute_url()


@register.filter
def job_result(job: JobModel) -> Optional[str]:
    queue = get_queue(job.queue_name)
    result = Result.fetch_latest(queue.connection, job.name)
    return result.type.name.capitalize() if result is not None else None


@register.filter
def job_scheduled_task(job: JobModel) -> Optional[str]:
    task = Task.objects.filter(id=job.scheduled_task_id).first()
    return task.name if task is not None else None


@register.filter
def job_status(job: JobModel):
    result = job.status
    return result.capitalize()


@register.filter
def job_runtime(job: JobModel):
    ended_at = job.ended_at
    if ended_at:
        runtime = job.ended_at - job.started_at
        return f"{int(runtime.microseconds / 1000)}ms"
    elif job.started_at:
        return "Still running"
    else:
        return "-"


@register.filter
def job_scheduled_time(job: JobModel, queue: Queue):
    return queue.scheduled_job_registry.get_scheduled_time(job.name)



from django.core.exceptions import ValidationError

from scheduler import settings
from scheduler.helpers.queues import get_queue
from scheduler.models import TaskType
from scheduler.redis_models import JobModel
from scheduler.tests.test_task_types.test_task_model import BaseTestCases
from scheduler.tests.testtools import task_factory
from scheduler.worker import create_worker


class TestCronTask(BaseTestCases.TestBaseTask):
    task_type = TaskType.CRON

    def setUp(self) -> None:
        super().setUp()
        self.queue_name = settings.get_queue_names()[0]

    def test_clean(self):
        task = task_factory(self.task_type)
        task.cron_string = "* * * * *"
        task.queue = self.queue_name
        task.callable = "scheduler.tests.jobs.test_job"
        self.assertIsNone(task.clean())

    def test_clean_cron_string_invalid(self):
        task = task_factory(self.task_type)
        task.cron_string = "not-a-cron-string"
        task.queue = self.queue_name
        task.callable = "scheduler.tests.jobs.test_job"
        with self.assertRaises(ValidationError):
            task.clean_cron_string()

    def test_check_rescheduled_after_execution(self):
        task = task_factory(self.task_type)
        queue = task.rqueue
        first_run_id = task.job_name
        entry = JobModel.get(first_run_id, connection=queue.connection)
        self.assertIsNotNone(entry)
        queue.run_sync(entry)
        task.refresh_from_db()
        self.assertEqual(task.failed_runs, 0)
        self.assertIsNone(task.last_failed_run)
        self.assertEqual(task.successful_runs, 1)
        self.assertIsNotNone(task.last_successful_run)
        self.assertTrue(task.is_scheduled())
        self.assertNotEqual(task.job_name, first_run_id)

    def test_check_rescheduled_after_failed_execution(self):
        task = task_factory(self.task_type, callable_name="scheduler.tests.jobs.failing_job")
        queue = task.rqueue
        first_run_id = task.job_name
        entry = JobModel.get(first_run_id, connection=queue.connection)
        queue.run_sync(entry)
        task.refresh_from_db()
        self.assertEqual(task.failed_runs, 1)
        self.assertIsNotNone(task.last_failed_run)
        self.assertEqual(task.successful_runs, 0)
        self.assertIsNone(task.last_successful_run)
        self.assertTrue(task.is_scheduled())
        self.assertNotEqual(task.job_name, first_run_id)

    def test_cron_task_enqueuing_jobs(self):
        queue = get_queue()
        prev_queued = queue.scheduled_job_registry.count(connection=queue.connection)
        prev_finished = queue.finished_job_registry.count(connection=queue.connection)

        task = task_factory(self.task_type, callable_name="scheduler.tests.jobs.enqueue_jobs")
        self.assertEqual(prev_queued + 1, queue.scheduled_job_registry.count(connection=queue.connection))
        first_run_id = task.job_name
        entry = JobModel.get(first_run_id, connection=queue.connection)
        queue.run_sync(entry)
        self.assertEqual(20, len(queue.queued_job_registry))
        self.assertEqual(prev_finished + 1, queue.finished_job_registry.count(connection=queue.connection))
        worker = create_worker("default", fork_job_execution=False, burst=True)
        worker.work()
        self.assertEqual(prev_finished + 21, queue.finished_job_registry.count(connection=queue.connection))
        worker.refresh(update_queues=True)
        self.assertEqual(20, worker._model.successful_job_count)
        self.assertEqual(0, worker._model.failed_job_count)



# Generated by Django 5.1.6 on 2025-02-05 15:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("scheduler", "0019_task_crontask_new_task_id_repeatabletask_new_task_id_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="repeatabletask",
            name="new_task_id",
        ),
        migrations.RemoveField(
            model_name="scheduledtask",
            name="new_task_id",
        ),
        migrations.DeleteModel(
            name="CronTask",
        ),
        migrations.DeleteModel(
            name="RepeatableTask",
        ),
        migrations.DeleteModel(
            name="ScheduledTask",
        ),
    ]



import contextlib
import errno
import math
import os
import random
import signal
import socket
import sys
import threading
import time
import traceback
import warnings
from datetime import timedelta
from enum import Enum
from random import shuffle
from resource import struct_rusage
from types import FrameType
from typing import List, Optional, Tuple, Any, Iterable

import scheduler
from scheduler.helpers.queues import get_queue
from scheduler.redis_models import WorkerModel, JobModel, JobStatus, DequeueTimeout
from scheduler.settings import SCHEDULER_CONFIG, logger, get_queue_configuration
from scheduler.types import Broker, Self
from scheduler.types import (
    ConnectionType,
    TimeoutErrorTypes,
    ConnectionErrorTypes,
    WatchErrorTypes,
    ResponseErrorTypes,
)
from .commands import WorkerCommandsChannelListener
from .scheduler import WorkerScheduler, SchedulerStatus
from ..redis_models.lock import QueueLock
from ..redis_models.worker import WorkerStatus

try:
    from signal import SIGKILL
except ImportError:
    from signal import SIGTERM as SIGKILL

from contextlib import suppress

from scheduler.helpers.queues import Queue, perform_job
from scheduler.timeouts import (
    JobExecutionMonitorTimeoutException,
    JobTimeoutException,
)
from scheduler.helpers.utils import utcnow, current_timestamp

try:
    from setproctitle import setproctitle as setprocname
except ImportError:

    def setprocname(*args, **kwargs):  # noqa
        pass


class StopRequested(Exception):
    pass


class WorkerNotFound(Exception):
    pass


class QueueConnectionDiscrepancyError(Exception):
    pass


_signames = dict(
    (getattr(signal, signame), signame) for signame in dir(signal) if signame.startswith("SIG") and "_" not in signame
)


def signal_name(signum):
    try:
        return signal.Signals(signum).name
    except KeyError:
        return "SIG_UNKNOWN"
    except ValueError:
        return "SIG_UNKNOWN"


class DequeueStrategy(str, Enum):
    DEFAULT = "default"
    ROUND_ROBIN = "round_robin"
    RANDOM = "random"


class Worker:
    queue_class = Queue

    # factor to increase connection_wait_time in case of continuous connection failures.
    exponential_backoff_factor = 2.0
    # Max Wait time (in seconds) after which exponential_backoff_factor won't be applicable.
    max_connection_wait_time = 60.0

    @classmethod
    def from_model(cls, model: WorkerModel) -> Self:
        connection = get_queue(model.queue_names[0]).connection
        res = cls(
            queues=[get_queue(queue_name) for queue_name in model.queue_names],
            name=model.name,
            connection=connection,
            with_scheduler=False,
            model=model,
        )
        return res

    def __init__(
        self,
        queues,
        name: str,
        connection: Optional[ConnectionType] = None,
        maintenance_interval: int = SCHEDULER_CONFIG.DEFAULT_MAINTENANCE_TASK_INTERVAL,
        job_monitoring_interval=SCHEDULER_CONFIG.DEFAULT_JOB_MONITORING_INTERVAL,
        dequeue_strategy: DequeueStrategy = DequeueStrategy.DEFAULT,
        disable_default_exception_handler: bool = False,
        fork_job_execution: bool = True,
        with_scheduler: bool = True,
        burst: bool = False,
        model: Optional[WorkerModel] = None,
    ):  # noqa
        self.fork_job_execution = fork_job_execution
        self.job_monitoring_interval = job_monitoring_interval
        self.maintenance_interval = maintenance_interval

        connection = self._set_connection(connection)
        self.connection = connection

        self.queues = [
            (Queue(name=q, connection=connection) if isinstance(q, str) else q) for q in _ensure_list(queues)
        ]
        self.name: str = name
        self._validate_name_uniqueness()
        self._ordered_queues = self.queues[:]

        self._is_job_execution_process: bool = False

        self.scheduler: Optional[WorkerScheduler] = None
        self._command_listener = WorkerCommandsChannelListener(connection, self.name)
        self._dequeue_strategy = dequeue_strategy

        self.disable_default_exception_handler = disable_default_exception_handler
        self.with_scheduler = with_scheduler
        self.burst = burst
        self._model = (
            model
            if model is not None
            else WorkerModel(
                name=self.name,
                queue_names=[queue.name for queue in self.queues],
                birth=None,
                last_heartbeat=None,
                pid=os.getpid(),
                hostname=socket.gethostname(),
                ip_address=_get_ip_address_from_connection(self.connection, self.name),
                version=scheduler.__version__,
                python_version=sys.version,
                state=WorkerStatus.CREATED,
            )
        )
        self._model.save(self.connection)

    @property
    def _pid(self) -> int:
        return self._model.pid

    def should_run_maintenance_tasks(self):
        """Maintenance tasks should run on first startup or every 10 minutes."""
        if self._model.last_cleaned_at is None:
            return True
        if (utcnow() - self._model.last_cleaned_at) > timedelta(seconds=self.maintenance_interval):
            return True
        return False

    def _set_connection(self, connection: ConnectionType) -> ConnectionType:
        """Configures the Broker connection to have a socket timeout.
        This should timouet the connection in case any specific command hangs at any given time (eg. BLPOP).
        If the connection provided already has a `socket_timeout` defined, skips.

        :param connection: Broker connection to configure.
        """
        current_socket_timeout = connection.connection_pool.connection_kwargs.get("socket_timeout")
        if current_socket_timeout is None:
            timeout_config = {"socket_timeout": self.connection_timeout}
            connection.connection_pool.connection_kwargs.update(timeout_config)
        return connection

    def clean_registries(self):
        """Runs maintenance jobs on each Queue's registries."""
        for queue in self.queues:
            # If there are multiple workers running, we only want 1 worker
            # to run clean_registries().
            queue_lock = QueueLock(self.name)
            if queue_lock.acquire(1, expire=899, connection=self.connection):
                logger.info(f"[Worker {self.name}/{self._pid}]: Cleaning registries for queue: {queue.name}")
                queue.clean_registries()
                WorkerModel.cleanup(queue.connection, queue.name)
                queue_lock.release(self.connection)
        self._model.last_cleaned_at = utcnow()

    def _install_signal_handlers(self) -> None:
        """Installs signal handlers for handling SIGINT and SIGTERM gracefully."""
        if threading.current_thread() is not threading.main_thread():
            logger.warning(
                f"[Worker {self.name}/{self._pid}]: Running in a thread, skipping signal handlers installation"
            )
            return
        signal.signal(signal.SIGINT, self.request_stop)
        signal.signal(signal.SIGTERM, self.request_stop)

    def work(
        self,
        max_jobs: Optional[int] = None,
        max_idle_time: Optional[int] = None,
    ) -> bool:
        """Starts the work loop.

        Pops and performs all jobs on the current list of queues.  When all
        queues are empty, block and wait for new jobs to arrive on any of the
        queues, unless `burst` mode is enabled.
        If `max_idle_time` is provided, worker will die when it's idle for more than the provided value.

        The return value indicates whether any jobs were processed.

        :param max_jobs: Max number of jobs. Defaults to None.
        :param max_idle_time: Max seconds for a worker to be idle. Defaults to None.
        :return: Whether any jobs were processed.
        """
        self.bootstrap()

        self._install_signal_handlers()
        try:
            while True:
                self.refresh()
                self._check_for_suspension(self.burst)

                if self.should_run_maintenance_tasks():
                    self.run_maintenance_tasks()

                if self._model.shutdown_requested_date:
                    logger.info(f"[Worker {self.name}/{self._pid}]: stopping on request")
                    break

                timeout = None if self.burst else (SCHEDULER_CONFIG.DEFAULT_WORKER_TTL - 15)
                job, queue = self.dequeue_job_and_maintain_ttl(timeout, max_idle_time)
                if job is None:
                    if self.burst:
                        logger.info(f"[Worker {self.name}/{self._pid}]: done, quitting")
                        break
                    elif max_idle_time is not None:
                        logger.info(f"[Worker {self.name}/{self._pid}]: idle for {max_idle_time} seconds, quitting")
                        break
                    continue

                self.execute_job(job, queue)

                self.refresh()
                with self.connection.pipeline() as pipeline:
                    self._model.heartbeat(pipeline)
                    self._model.save(pipeline)
                    pipeline.execute()
                if max_jobs is not None and self._model.completed_jobs >= max_jobs:
                    logger.info(
                        f"[Worker {self.name}/{self._pid}]: finished executing {self._model.completed_jobs} jobs, quitting"
                    )
                    break
            return self._model.completed_jobs > 0

        except TimeoutErrorTypes:
            logger.error(f"[Worker {self.name}/{self._pid}]: Redis connection timeout, quitting...")
        except StopRequested:
            logger.info(f"[Worker {self.name}/{self._pid}]: Worker was requested to stop, quitting")
            pass
        except SystemExit:  # Cold shutdown detected
            raise
        except Exception:
            logger.error(f"[Worker {self.name}/{self._pid}]: found an unhandled exception, quitting...", exc_info=True)
        finally:
            self.teardown()

    def handle_job_failure(self, job: JobModel, queue: Queue, exc_string=""):
        """
        Handles the failure or an executing job by:
            1. Setting the job status to failed
            2. Removing the job from active_job_registry
            3. Setting the workers current job to None
            4. Add the job to FailedJobRegistry
        `save_exc_to_job` should only be used for testing purposes
        """
        logger.debug(f"[Worker {self.name}/{self._pid}]: Handling failed execution of job {job.name}")
        # check whether a job was stopped intentionally and set the job status appropriately if it was this job.

        new_job_status = (
            JobStatus.STOPPED
            if self._model.get_field("stopped_job_name", self.connection) == job.name
            else JobStatus.FAILED
        )
        self._model.current_job_name = None
        with self.connection.pipeline() as pipeline:
            if new_job_status == JobStatus.STOPPED:
                logger.debug(f"[Worker {self.name}/{self._pid}]: Job was stopped, setting status to STOPPED")
            else:
                logger.debug(f"[Worker {self.name}/{self._pid}]: Job has failed, setting status to FAILED")

            queue.active_job_registry.delete(connection=pipeline, job_name=job.name)

            if not self.disable_default_exception_handler:
                queue.job_handle_failure(new_job_status, job, exc_string, connection=pipeline)
                with suppress(ConnectionErrorTypes):
                    pipeline.execute()

            self._model.current_job_working_time = 0
            if job.status == JobStatus.FAILED:
                self._model.failed_job_count += 1
                self._model.completed_jobs += 1
            if job.started_at and job.ended_at:
                self._model.total_working_time_ms += (job.ended_at - job.started_at).microseconds / 1000.0
            self._model.save(connection=self.connection)

            try:
                pipeline.execute()
            except Exception:
                # Ensure that custom exception handlers are called even if the Broker is down
                pass

    def bootstrap(self):
        """Bootstraps the worker.
        Runs the basic tasks that should run when the worker actually starts working.
        Used so that new workers can focus on the work loop implementation rather
        than the full bootstraping process.
        """
        self.worker_start()
        logger.info(f"[Worker {self.name}/{self._pid}]: Worker {self.name} started with PID {os.getpid()}")
        self._command_listener.start()
        if self.with_scheduler:
            self.scheduler = WorkerScheduler(self.queues, worker_name=self.name, connection=self.connection)
            self.scheduler.start()
            self._model.has_scheduler = True
            self._model.save(connection=self.connection)
        if self.with_scheduler and self.burst:
            self.scheduler.request_stop_and_wait()
            self._model.has_scheduler = False
            self._model.save(connection=self.connection)
        qnames = [queue.name for queue in self.queues]
        logger.info(f"""[Worker {self.name}/{self._pid}]: Listening to queues {", ".join(qnames)}...""")

    def _check_for_suspension(self, burst: bool) -> None:
        """Check to see if workers have been suspended by `rq suspend`"""
        before_state = None
        notified = False
        while self._model.is_suspended:
            if burst:
                logger.info(
                    f"[Worker {self.name}/{self._pid}]: Suspended in burst mode, exiting, "
                    f"Note: There could still be unfinished jobs on the queue"
                )
                raise StopRequested()

            if not notified:
                logger.info(f"[Worker {self.name}/{self._pid}]: Worker suspended, trigger ResumeCommand")
                before_state = self._model.state
                self._model.set_field("state", WorkerStatus.SUSPENDED, connection=self.connection)
                notified = True
            time.sleep(1)

        if before_state:
            self._model.set_field("state", before_state, connection=self.connection)

    def run_maintenance_tasks(self):
        """Runs periodic maintenance tasks, these include:
        1. Check if scheduler should be started.
        2. Cleaning registries
        """
        self.clean_registries()
        if not self.with_scheduler:
            return
        if self.scheduler is None:
            self.scheduler = WorkerScheduler(self.queues, worker_name=self.name, connection=self.connection)
        if self.scheduler.status == SchedulerStatus.STOPPED:
            self.scheduler.start()
            self._model.has_scheduler = True
            self._model.save(connection=self.connection)
        if self.burst:
            self.scheduler.request_stop_and_wait()
            self._model.has_scheduler = False
            self._model.save(connection=self.connection)

    def dequeue_job_and_maintain_ttl(
        self, timeout: Optional[int], max_idle_time: Optional[int] = None
    ) -> Tuple[JobModel, Queue]:
        """Dequeues a job while maintaining the TTL.
        :param timeout: The timeout for the dequeue operation.
        :param max_idle_time: The maximum idle time for the worker.
        :returns: A tuple with the job and the queue.
        """
        qnames = ",".join([queue.name for queue in self.queues])

        self._model.set_field("state", WorkerStatus.IDLE, connection=self.connection)
        self.procline(f"Listening on {qnames}")
        logger.debug(f"[Worker {self.name}/{self._pid}]: listening on {qnames}...")
        connection_wait_time = 1.0
        idle_since = utcnow()
        idle_time_left = max_idle_time
        job, queue = None, None
        while True:
            try:
                self._model.heartbeat(self.connection)

                if self.should_run_maintenance_tasks():
                    self.run_maintenance_tasks()

                if timeout is not None and idle_time_left is not None:
                    timeout = min(timeout, idle_time_left)

                logger.debug(
                    f"[Worker {self.name}/{self._pid}]: Fetching jobs on queues {qnames} and timeout {timeout}"
                )
                job, queue = Queue.dequeue_any(self._ordered_queues, timeout, connection=self.connection)
                if job is not None:
                    self.reorder_queues(reference_queue=queue)
                    logger.info(f"[Worker {self.name}/{self._pid}]: Popped job `{job.name}` from `{queue.name}`")
                break
            except DequeueTimeout:
                if max_idle_time is not None:
                    idle_for = (utcnow() - idle_since).total_seconds()
                    idle_time_left = math.ceil(max_idle_time - idle_for)
                    if idle_time_left <= 0:
                        break
            except ConnectionErrorTypes as conn_err:
                logger.error(
                    f"[Worker {self.name}/{self._pid}]: Could not connect to Broker: {conn_err} Retrying in {connection_wait_time} seconds..."
                )
                time.sleep(connection_wait_time)
                connection_wait_time *= self.exponential_backoff_factor
                connection_wait_time = min(connection_wait_time, self.max_connection_wait_time)

        self._model.heartbeat(self.connection)
        return job, queue

    @property
    def connection_timeout(self) -> int:
        return SCHEDULER_CONFIG.DEFAULT_WORKER_TTL - 5

    def procline(self, message):
        """Changes the current procname for the process.

        This can be used to make `ps -ef` output more readable.
        """
        setprocname(f"{self._model._key}: {message}")

    def _validate_name_uniqueness(self):
        """Validates that the worker name is unique."""
        worker_model = WorkerModel.get(self.name, connection=self.connection)
        if worker_model is not None and worker_model.death is None:
            raise ValueError(f"There exists an active worker named {self.name!r} already")

    def worker_start(self):
        """Registers its own birth."""
        logger.debug(f"[Worker {self.name}/{self._pid}]: Registering birth")
        now = utcnow()
        self._model.birth = now
        self._model.last_heartbeat = now
        self._model.state = WorkerStatus.STARTED
        self._model.save(self.connection)

    def kill_job_execution_process(self, sig: signal.Signals = SIGKILL):
        """Kill the job execution process but catch "No such process" error has the job execution process could already
        be dead.

        :param sig: Optional, Defaults to SIGKILL.
        """
        try:
            os.killpg(os.getpgid(self._model.job_execution_process_pid), sig)
            logger.info(
                f"[Worker {self.name}/{self._pid}]: Killed job execution process pid {self._model.job_execution_process_pid}"
            )
        except OSError as e:
            if e.errno == errno.ESRCH:
                # "No such process" is fine with us
                logger.debug("[Worker {self.name}/{self._pid}]: Job execution process already dead")
            else:
                raise

    def wait_for_job_execution_process(self) -> Tuple[Optional[int], Optional[int], Optional[struct_rusage]]:
        """Waits for the job execution process to complete.
        Uses `0` as argument as to include "any child in the process group of the current process".
        """
        pid = stat = rusage = None
        with contextlib.suppress(ChildProcessError):  # ChildProcessError: [Errno 10] No child processes
            pid, stat, rusage = os.wait4(self._model.job_execution_process_pid, 0)
        return pid, stat, rusage

    def request_force_stop(self, signum: int, frame: Optional[FrameType]):
        """Terminates the application (cold shutdown).

        :param signum: Signal number
        :param frame: Frame
        :raises SystemExit: SystemExit
        """
        # When a worker is run through a worker pool, it may receive duplicate signals.
        # One is sent by the pool when it calls `pool.stop_worker()` and another is sent by the OS
        # when a user hits Ctrl+C. In this case, if we receive the second signal within 1 second, we ignore it.
        shutdown_date = self._model.shutdown_requested_date
        if shutdown_date is not None and (utcnow() - shutdown_date) < timedelta(seconds=1):
            logger.debug(
                f"[Worker {self.name}/{self._pid}]: Shutdown signal ignored, received twice in less than 1 second"
            )
            return

        logger.warning(f"[Worker {self.name}/{self._pid}]: Cold shut down")

        # Take down the job execution process with the worker
        if self._model.job_execution_process_pid:
            logger.debug(
                f"[Worker {self.name}/{self._pid}]: Taking down job execution process {self._model.job_execution_process_pid} with me"
            )
            self.kill_job_execution_process()
            self.wait_for_job_execution_process()
        raise SystemExit()

    def request_stop(self, signum: int, frame: Optional[FrameType]) -> None:
        """Stops the current worker loop but waits for child processes to end gracefully (warm shutdown).
        :param signum: Signal number
        :param frame: Frame
        """
        logger.debug(f"[Worker {self.name}/{self._pid}]: Got signal {signal_name(signum)}")
        self._model.set_field("shutdown_requested_date", utcnow(), self.connection)

        signal.signal(signal.SIGINT, self.request_force_stop)
        signal.signal(signal.SIGTERM, self.request_force_stop)

        logger.info(f"[Worker {self.name}/{self._pid}]: warm shut down requested")

        self.stop_scheduler()
        # If shutdown is requested in the middle of a job, wait until finish before shutting down and save the request.
        if self._model.state == WorkerStatus.BUSY:
            self._model.set_field("shutdown_requested_date", utcnow(), connection=self.connection)

            logger.debug(
                f"[Worker {self.name}/{self._pid}]: Stopping after current job execution process is finished. "
                f"Press Ctrl+C again for a cold shutdown."
            )
        else:
            raise StopRequested()

    def reorder_queues(self, reference_queue: Queue):
        """Reorder the queues according to the strategy.
        As this can be defined both in the `Worker` initialization or in the `work` method,
        it doesn't take the strategy directly, but rather uses the private `_dequeue_strategy` attribute.

        :param reference_queue: The queues to reorder
        """
        if self._dequeue_strategy is None:
            self._dequeue_strategy = DequeueStrategy.DEFAULT

        if self._dequeue_strategy not in [e.value for e in DequeueStrategy]:
            raise ValueError(
                f"""[Worker {self.name}/{self._pid}]: Dequeue strategy should be one of {", ".join([e.value for e in DequeueStrategy])}"""
            )
        if self._dequeue_strategy == DequeueStrategy.DEFAULT:
            return
        if self._dequeue_strategy == DequeueStrategy.ROUND_ROBIN:
            pos = self._ordered_queues.index(reference_queue)
            self._ordered_queues = self._ordered_queues[pos + 1 :] + self._ordered_queues[: pos + 1]
            return
        if self._dequeue_strategy == DequeueStrategy.RANDOM:
            shuffle(self._ordered_queues)
            return

    def teardown(self) -> None:
        if self._is_job_execution_process:
            return
        self.stop_scheduler()
        self._command_listener.stop()
        self._model.delete(self.connection)

    def stop_scheduler(self):
        """Stop the scheduler thread.
        Will send the kill signal to the scheduler process,
        if there's an OSError, just passes and `join()`'s the scheduler process, waiting for the process to finish.
        """
        if self.scheduler is None:
            return
        logger.info(f"[Worker {self.name}/{self._pid}]: Stopping scheduler thread {self.scheduler.pid}")
        self.scheduler.request_stop_and_wait()
        logger.debug(f"[Worker {self.name}/{self._pid}]: Scheduler thread stopped")
        self.scheduler = None

    def refresh(self, update_queues: bool = False):
        """Refreshes the worker data.
        It will get the data from the datastore and update the Worker's attributes
        """
        self._model = WorkerModel.get(self.name, connection=self.connection)
        if self._model is None:
            msg = f"[Worker {self.name}/{self._pid}]: Worker broker record for {self.name} not found, quitting..."
            logger.error(msg)
            raise WorkerNotFound(msg)
        if update_queues:
            self.queues = [Queue(name=queue_name, connection=self.connection) for queue_name in self._model.queue_names]

    def fork_job_execution_process(self, job: JobModel, queue: Queue) -> None:
        """Spawns a job execution process to perform the actual work and passes it a job.
        This is where the `fork()` actually happens.

        :param job: The job to be executed
        :param queue: The queue from which the job was dequeued
        """
        child_pid = os.fork()
        os.environ["SCHEDULER_WORKER_NAME"] = self.name
        os.environ["SCHEDULER_JOB_NAME"] = job.name
        if child_pid == 0:  # Child process/Job executor process to run the job
            os.setsid()
            self._model.job_execution_process_pid = os.getpid()
            self._model.save(connection=self.connection)
            self.execute_in_separate_process(job, queue)
            os._exit(0)  # just in case
        else:  # Parent worker process
            logger.debug(
                f"[Worker {self.name}/{self._pid}]: Forking job execution process, job_execution_process_pid={child_pid}"
            )
            self._model.job_execution_process_pid = child_pid
            self._model.save(connection=self.connection)
            self.procline(f"Forked {child_pid} at {time.time()}")

    def get_heartbeat_ttl(self, job: JobModel) -> int:
        """Get's the TTL for the next heartbeat.
        :param job: The Job
        :return: The heartbeat TTL
        """
        if job.timeout and job.timeout > 0:
            remaining_execution_time = int(job.timeout - self._model.current_job_working_time)
            return min(remaining_execution_time, self.job_monitoring_interval) + 60
        else:
            return self.job_monitoring_interval + 60

    def monitor_job_execution_process(self, job: JobModel, queue: Queue) -> None:
        """The worker will monitor the job execution process and make sure that it either executes successfully or the
        status of the job is set to failed

        :param job: The Job
        :param queue: The Queue
        """
        retpid = ret_val = rusage = None
        job.started_at = utcnow()
        while True:
            try:
                with SCHEDULER_CONFIG.DEATH_PENALTY_CLASS(
                    self.job_monitoring_interval, JobExecutionMonitorTimeoutException
                ):
                    retpid, ret_val, rusage = self.wait_for_job_execution_process()
                break
            except JobExecutionMonitorTimeoutException:
                # job execution process has not exited yet and is still running. Send a heartbeat to keep the worker alive.
                self._model.set_current_job_working_time((utcnow() - job.started_at).total_seconds(), self.connection)

                # Kill the job from this side if something is really wrong (interpreter lock/etc).
                if job.timeout != -1 and self._model.current_job_working_time > (job.timeout + 60):
                    self._model.heartbeat(self.connection, self.job_monitoring_interval + 60)
                    self.kill_job_execution_process()
                    self.wait_for_job_execution_process()
                    break

                self.maintain_heartbeats(job, queue)

            except OSError as e:
                # In case we encountered an OSError due to EINTR (which is
                # caused by a SIGINT or SIGTERM signal during
                # os.waitpid()), we simply ignore it and enter the next
                # iteration of the loop, waiting for the child to end.  In
                # any other case, this is some other unexpected OS error,
                # which we don't want to catch, so we re-raise those ones.
                if e.errno != errno.EINTR:
                    raise
                # Send a heartbeat to keep the worker alive.
                self._model.heartbeat(self.connection)

        self._model = WorkerModel.get(self.name, connection=self.connection)
        self._model.current_job_working_time = 0
        self._model.save(connection=self.connection)
        if ret_val == os.EX_OK:  # The process exited normally.
            return

        job_status = job.get_status(self.connection)
        stopped_job_name = self._model.get_field("stopped_job_name", self.connection)

        if job_status is None:  # Job completed and its ttl has expired
            logger.debug(f"[Worker {self.name}/{self._pid}]: Job status is None, completed and expired?")
            return
        elif stopped_job_name == job.name:  # job execution process killed deliberately
            logger.warning(f"[Worker {self.name}/{self._pid}]: Job stopped by user, moving job to failed-jobs-registry")
            if job.stopped_callback:
                job.stopped_callback()
            self.handle_job_failure(
                job, queue=queue, exc_string="Job stopped by user, job execution process terminated."
            )
        elif job_status not in [JobStatus.FINISHED, JobStatus.FAILED]:
            if not job.ended_at:
                job.ended_at = utcnow()

            # Unhandled failure: move the job to the failed queue
            signal_msg = f" (signal {os.WTERMSIG(ret_val)})" if ret_val and os.WIFSIGNALED(ret_val) else ""
            exc_string = f"job-execution-process terminated unexpectedly; waitpid returned {ret_val}{signal_msg}; "
            logger.warning(
                f"[Worker {self.name}/{self._pid}]: Moving job to {queue.name}/failed-job-registry ({exc_string})"
            )

            self.handle_job_failure(job, queue=queue, exc_string=exc_string)

    def execute_job(self, job: JobModel, queue: Queue):
        """Spawns a job execution process to perform the actual work and passes it a job.
        The worker will wait for the job execution process and make sure it executes within the given timeout bounds, or
        will end the job execution process with SIGALRM.
        """
        if self.fork_job_execution:
            self._model.set_field("state", WorkerStatus.BUSY, connection=self.connection)
            self.fork_job_execution_process(job, queue)
            self.monitor_job_execution_process(job, queue)
            self._model.set_field("state", WorkerStatus.IDLE, connection=self.connection)
        else:
            self._model.set_field("state", WorkerStatus.BUSY, connection=self.connection)
            self.perform_job(job, queue)
            self._model.set_field("state", WorkerStatus.IDLE, connection=self.connection)

    def maintain_heartbeats(self, job: JobModel, queue: Queue):
        """Updates worker and job's last heartbeat field."""
        with self.connection.pipeline() as pipeline:
            self._model.heartbeat(pipeline, self.job_monitoring_interval + 60)
            ttl = self.get_heartbeat_ttl(job)

            queue.active_job_registry.add(pipeline, self.name, current_timestamp() + ttl, update_existing_only=False)
            results = pipeline.execute()
            if results[2] == 1:
                job.delete(self.connection)

    def execute_in_separate_process(self, job: JobModel, queue: Queue):
        """This is the entry point of the newly spawned job execution process.
        After fork()'ing, assure we are generating random sequences that are different from the worker.

        os._exit() is the way to exit from child processes after a fork(), in contrast to the regular sys.exit()
        """
        random.seed()
        self.setup_job_execution_process_signals()
        self._is_job_execution_process = True
        try:
            self.perform_job(job, queue)
        except:  # noqa
            os._exit(1)
        os._exit(0)

    def setup_job_execution_process_signals(self):
        """Setup signal handing for the newly spawned job execution process

        Always ignore Ctrl+C in the job execution process, as it might abort the currently running job.

        The main worker catches the Ctrl+C and requests graceful shutdown after the current work is done.
        When cold shutdown is requested, it kills the current job anyway.
        """
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

    def worker_before_execution(self, job: JobModel, connection: ConnectionType) -> None:
        logger.debug(f"[Worker {self.name}/{self._pid}]: Preparing for execution of job: `{job.name}`")
        current_pid = os.getpid()
        self._model.current_job_name = job.name
        self._model.current_job_working_time = 0
        self._model.job_execution_process_pid = current_pid
        heartbeat_ttl = self.get_heartbeat_ttl(job)
        self._model.heartbeat(self.connection, heartbeat_ttl)
        self.procline(
            f"[Worker {self.name}/{self._pid}]: Processing {job.func_name} from {job.queue_name} since {time.time()}"
        )
        self._model.save(connection=connection)

    def handle_job_success(self, job: JobModel, return_value: Any, queue: Queue):
        """Handles the successful execution of certain job.
        It will remove the job from the `active_job_registry`, adding it to the `SuccessfulJobRegistry`,
        and run a few maintenance tasks including:
            - Resting the current job name
            - Enqueue dependents
            - Incrementing the job count and working time
            - Handling of the job successful execution

        Runs within a loop with the `watch` method so that protects interactions with dependents keys.

        :param job: The job that was successful.
        :param queue: The queue
        """
        logger.debug(f"[Worker {self.name}/{self._pid}]: Handling successful execution of job {job.name}")
        with self.connection.pipeline() as pipeline:
            while True:
                try:
                    queue.job_handle_success(
                        job,
                        result=return_value,
                        job_info_ttl=job.job_info_ttl,
                        result_ttl=job.success_ttl,
                        connection=pipeline,
                    )
                    self._model.current_job_name = None
                    self._model.successful_job_count += 1
                    self._model.completed_jobs += 1
                    self._model.total_working_time_ms += (job.ended_at - job.started_at).microseconds / 1000.0
                    self._model.save(connection=self.connection)

                    job.expire(job.success_ttl, connection=pipeline)
                    logger.debug(f"[Worker {self.name}/{self._pid}]: Removing job {job.name} from active_job_registry")
                    pipeline.execute()
                    logger.debug(
                        f"[Worker {self.name}/{self._pid}]: Finished handling successful execution of job {job.name}"
                    )
                    break
                except WatchErrorTypes:
                    continue

    def perform_job(self, job: JobModel, queue: Queue) -> bool:
        """Performs the actual work of a job.
        Called from the process executing the job (forked job execution process).

        :param job: The job to perform
        :param queue: The queue the job was dequeued from
        :returns: True after finished.
        """
        logger.debug(f"[Worker {self.name}/{self._pid}]: Performing {job.name} code.")

        try:
            with self.connection.pipeline() as pipeline:
                self.worker_before_execution(job, connection=pipeline)
                job.prepare_for_execution(self.name, queue.active_job_registry, connection=pipeline)
                pipeline.execute()
            timeout = job.timeout or SCHEDULER_CONFIG.DEFAULT_JOB_TIMEOUT
            with SCHEDULER_CONFIG.DEATH_PENALTY_CLASS(timeout, JobTimeoutException, job_name=job.name):
                logger.debug(f"[Worker {self.name}/{self._pid}]: Performing job `{job.name}`...")
                rv = perform_job(job, self.connection)
                logger.debug(f"[Worker {self.name}/{self._pid}]: Finished performing job `{job.name}`")

            self.handle_job_success(job=job, return_value=rv, queue=queue)
        except:  # NOQA
            logger.debug(f"[Worker {self.name}/{self._pid}]: Job {job.name} raised an exception.")
            exc_info = sys.exc_info()
            exc_string = "".join(traceback.format_exception(*exc_info))

            self.handle_job_failure(job=job, exc_string=exc_string, queue=queue)
            self.handle_exception(job, *exc_info)
            return False

        logger.info(f"[Worker {self.name}/{self._pid}]: queue:{queue.name}/job:{job.name} performed.")
        logger.debug(f"[Worker {self.name}/{self._pid}]: job:{job.name} result: {str(rv)}")

        return True

    def handle_exception(self, job: JobModel, *exc_info):
        """Walks the exception handler stack to delegate exception handling.
        If the job cannot be deserialized, it will raise when func_name or
        the other properties are accessed, which will stop exceptions from
        being properly logged, so we guard against it here.
        """
        logger.debug(f"Handling exception caused while performing job:{job.name}.")
        exc_string = "".join(traceback.format_exception(*exc_info))

        extra = {
            "func": job.func_name,
            "arguments": job.args,
            "kwargs": job.kwargs,
            Queue: job.queue_name,
            "job_name": job.name,
        }
        func_name = job.func_name

        # func_name
        logger.error(
            f"[Worker {self.name}/{self._pid}]: exception raised while executing ({func_name})\n{exc_string}",
            extra=extra,
        )


class SimpleWorker(Worker):
    def execute_job(self, job: JobModel, queue: Queue):
        """Execute job in same thread/process, do not fork()"""
        self._model.set_field("state", WorkerStatus.BUSY, connection=self.connection)
        self.perform_job(job, queue)
        self._model.set_field("state", WorkerStatus.IDLE, connection=self.connection)


class RoundRobinWorker(Worker):
    """Modified version of Worker that dequeues jobs from the queues using a round-robin strategy."""

    def reorder_queues(self, reference_queue):
        pos = self._ordered_queues.index(reference_queue)
        self._ordered_queues = self._ordered_queues[pos + 1 :] + self._ordered_queues[: pos + 1]


class RandomWorker(Worker):
    """Modified version of Worker that dequeues jobs from the queues using a random strategy."""

    def reorder_queues(self, reference_queue):
        shuffle(self._ordered_queues)


def _get_ip_address_from_connection(connection: ConnectionType, client_name: str) -> str:
    try:
        connection.client_setname(client_name)
    except ResponseErrorTypes:
        warnings.warn("CLIENT SETNAME command not supported, setting ip_address to unknown", Warning)
        return "unknown"
    client_adresses = [client["addr"] for client in connection.client_list() if client["name"] == client_name]
    if len(client_adresses) > 0:
        return client_adresses[0]
    else:
        warnings.warn("CLIENT LIST command not supported, setting ip_address to unknown", Warning)
        return "unknown"


def _ensure_list(obj: Any) -> List:
    """When passed an iterable of objects, does nothing, otherwise, it returns a list with just that object in it.

    :param obj: The object to ensure is a list
    :return:
    """
    is_nonstring_iterable = isinstance(obj, Iterable) and not isinstance(obj, str)
    return obj if is_nonstring_iterable else [obj]


def _calc_worker_name(existing_worker_names) -> str:
    hostname = os.uname()[1]
    c = 1
    worker_name = f"{hostname}-worker.{c}"
    while worker_name in existing_worker_names:
        c += 1
        worker_name = f"{hostname}-worker.{c}"
    return worker_name


def get_queues(*queue_names: str) -> List[Queue]:
    """Return queue instances from specified queue names. All instances must use the same connection configuration."""

    queue_config = get_queue_configuration(queue_names[0])
    queues = [get_queue(queue_names[0])]
    # perform consistency checks while building return list
    for queue_name in queue_names[1:]:
        curr_queue_config = get_queue_configuration(queue_name)
        if not queue_config.same_connection_params(curr_queue_config):
            raise QueueConnectionDiscrepancyError(
                f'Queues must have the same broker connection. "{queue_name}" and "{queue_names[0]}" have different connection settings'
            )
        queue = get_queue(queue_name)
        queues.append(queue)

    return queues


def create_worker(*queue_names: str, **kwargs) -> Worker:
    """Returns a Django worker for all queues or specified ones."""
    queues = get_queues(*queue_names)
    existing_worker_names = WorkerModel.all_names(connection=queues[0].connection)
    kwargs.setdefault("fork_job_execution", SCHEDULER_CONFIG.BROKER != Broker.FAKEREDIS)
    if kwargs.get("name", None) is None:
        kwargs["name"] = _calc_worker_name(existing_worker_names)
    if kwargs["name"] in existing_worker_names:
        raise ValueError(f"Worker {kwargs['name']} already exists")
    kwargs["name"] = kwargs["name"].replace("/", ".")
    kwargs.setdefault("with_scheduler", False)
    worker = Worker(queues, connection=queues[0].connection, **kwargs)
    return worker



__all__ = [
    "Queue",
    "InvalidJobOperation",
    "get_queue",
    "get_all_workers",
    "perform_job",
]

from .getters import get_queue, get_all_workers
from .queue_logic import Queue, InvalidJobOperation, perform_job



from time import sleep

from django.test import tag
from django.urls import reverse

from scheduler.helpers.queues import get_queue
from scheduler.redis_models import JobStatus, JobModel, WorkerModel
from scheduler.tests.jobs import long_job
from .. import testtools
from ..test_views.base import BaseTestCase


@tag("multiprocess")
class MultiProcessTest(BaseTestCase):
    def test_cancel_job_after_it_started(self):
        # arrange
        queue = get_queue("django_tasks_scheduler_test")
        job = queue.create_and_enqueue_job(long_job)
        self.assertTrue(job.is_queued)
        process, worker_name = testtools.run_worker_in_process("django_tasks_scheduler_test")
        sleep(0.2)
        job = JobModel.get(job.name, connection=queue.connection)
        self.assertEqual(JobStatus.STARTED, job.status)
        # act
        res = self.client.post(reverse("job_detail_action", args=[job.name, "cancel"]), {"post": "yes"}, follow=True)

        # assert
        self.assertEqual(200, res.status_code)
        job = JobModel.get(job.name, connection=queue.connection)
        self.assertEqual(JobStatus.STOPPED, job.status)
        self.assertNotIn(job.name, queue.queued_job_registry.all())
        sleep(0.2)
        process.terminate()
        process.join(2)
        process.kill()
        worker_model = WorkerModel.get(worker_name, connection=queue.connection)
        self.assertEqual(0, worker_model.completed_jobs)
        self.assertEqual(0, worker_model.failed_job_count)
        self.assertEqual(0, worker_model.successful_job_count)
        self.assertIsNotNone(worker_model.shutdown_requested_date)



import dataclasses
from datetime import datetime
from enum import Enum
from typing import Optional, Any, ClassVar, List

from scheduler.helpers.utils import utcnow
from scheduler.redis_models.base import StreamModel, decode_dict
from scheduler.types import ConnectionType, Self


class ResultType(Enum):
    SUCCESSFUL = "successful"
    FAILED = "failed"
    STOPPED = "stopped"


@dataclasses.dataclass(slots=True, kw_only=True)
class Result(StreamModel):
    parent: str
    type: ResultType
    worker_name: str
    ttl: Optional[int] = 0
    name: Optional[str] = None
    created_at: datetime = dataclasses.field(default_factory=utcnow)
    return_value: Optional[Any] = None
    exc_string: Optional[str] = None

    _list_key: ClassVar[str] = ":job-results:"
    _children_key_template: ClassVar[str] = ":job-results:{}:"
    _element_key_template: ClassVar[str] = ":job-results:{}"

    @classmethod
    def create(
        cls,
        connection: ConnectionType,
        job_name: str,
        worker_name: str,
        _type: ResultType,
        ttl: int,
        return_value: Any = None,
        exc_string: Optional[str] = None,
    ) -> Self:
        result = cls(
            parent=job_name,
            ttl=ttl,
            type=_type,
            return_value=return_value,
            exc_string=exc_string,
            worker_name=worker_name,
        )
        result.save(connection)
        return result

    @classmethod
    def fetch_latest(cls, connection: ConnectionType, job_name: str) -> Optional["Result"]:
        """Returns the latest result for given job_name.

        :param connection: Broker connection.
        :param job_name: Job name.
        :return: Result instance or None if no result is available.
        """
        response: List[Any] = connection.xrevrange(cls._children_key_template.format(job_name), "+", "-", count=1)
        if not response:
            return None
        result_id, payload = response[0]
        res = cls.deserialize(decode_dict(payload, set()))
        return res



from typing import Optional, Any

from scheduler.types import ConnectionType


class KvLock(object):
    def __init__(self, name: str) -> None:
        self.name = name
        self.acquired = False

    @property
    def _locking_key(self) -> str:
        return f"_lock:{self.name}"

    def acquire(self, val: Any, connection: ConnectionType, expire: Optional[int] = None) -> bool:
        self.acquired = connection.set(self._locking_key, val, nx=True, ex=expire)
        return self.acquired

    def expire(self, connection: ConnectionType, expire: Optional[int] = None) -> bool:
        return connection.expire(self._locking_key, expire)

    def release(self, connection: ConnectionType):
        connection.delete(self._locking_key)

    def value(self, connection: ConnectionType) -> Any:
        return connection.get(self._locking_key)


class SchedulerLock(KvLock):
    def __init__(self, queue_name: str) -> None:
        super().__init__(f"lock:scheduler:{queue_name}")


class QueueLock(KvLock):
    def __init__(self, queue_name: str) -> None:
        super().__init__(f"queue:{queue_name}")



from django.core.management import call_command

from scheduler.helpers.queues import get_queue
from scheduler.tests import test_settings  # noqa
from scheduler.tests.jobs import failing_job
from scheduler.worker import create_worker
from scheduler.tests.test_views.base import BaseTestCase


class DeleteFailedExecutionsTest(BaseTestCase):
    def test_delete_failed_executions__delete_jobs(self):
        queue = get_queue("default")
        call_command("delete_failed_executions", queue="default")
        queue.create_and_enqueue_job(failing_job)
        self.assertEqual(1, len(queue.queued_job_registry))
        worker = create_worker("default", burst=True)
        worker.work()
        self.assertEqual(1, len(queue.failed_job_registry))
        call_command("delete_failed_executions", queue="default")
        self.assertEqual(0, len(queue.failed_job_registry))



import json
import sys
from io import StringIO

import yaml
from django.core.management import call_command
from django.test import TestCase, override_settings

from scheduler import settings
from scheduler.helpers.queues import get_queue


@override_settings(SCHEDULER_QUEUES=dict(default={"HOST": "localhost", "PORT": 6379, "DB": 0}))
class SchedulerStatsTest(TestCase):
    EXPECTED_OUTPUT = {
        "queues": [
            {
                "canceled_jobs": 0,
                "failed_jobs": 0,
                "finished_jobs": 0,
                "name": "default",
                "oldest_job_timestamp": None,
                "queued_jobs": 0,
                "scheduled_jobs": 0,
                "scheduler_pid": None,
                "started_jobs": 0,
                "workers": 0,
            }
        ]
    }
    OLD_QUEUES = None

    def setUp(self):
        super(SchedulerStatsTest, self).setUp()
        SchedulerStatsTest.OLD_QUEUES = settings._QUEUES
        settings._QUEUES = dict()
        settings.conf_settings()
        get_queue("default").connection.flushall()

    def tearDown(self):
        super(SchedulerStatsTest, self).tearDown()
        settings._QUEUES = SchedulerStatsTest.OLD_QUEUES

    def test_scheduler_stats__json_output(self):
        test_stdout = StringIO()
        sys.stdout = test_stdout
        # act
        call_command("scheduler_stats", "-j")
        # assert
        res = test_stdout.getvalue()
        self.assertEqual(json.loads(res), SchedulerStatsTest.EXPECTED_OUTPUT)

    def test_scheduler_stats__yaml_output(self):
        # arrange
        test_stdout = StringIO()
        sys.stdout = test_stdout
        # act
        call_command("scheduler_stats", "-y")
        # assert
        res = test_stdout.getvalue()
        self.assertEqual(yaml.load(res, yaml.SafeLoader), SchedulerStatsTest.EXPECTED_OUTPUT)

    def test_scheduler_stats__plain_text_output(self):
        test_stdout = StringIO()
        sys.stdout = test_stdout
        # act
        call_command("scheduler_stats", "--no-color")
        # assert
        res = test_stdout.getvalue()
        self.assertEqual(
            res,
            """
Django-Scheduler CLI Dashboard

--------------------------------------------------------------------------------
| Name             |    Queued |    Active |  Finished |  Canceled |   Workers |
--------------------------------------------------------------------------------
| default          |         0 |         0 |         0 |         0 |         0 |
--------------------------------------------------------------------------------
""",
        )

    def test_scheduler_stats__bad_args(self):
        # arrange
        sys.stderr = StringIO()
        sys.stdout = StringIO()
        # act
        with self.assertRaises(SystemExit):
            call_command("scheduler_stats", "-y", "-j")
        # assert
        res = sys.stdout.getvalue()
        self.assertEqual(res, """""")
        err = sys.stderr.getvalue()
        self.assertEqual(err, """Aborting. Cannot output as both json and yaml\n""")



from typing import Set

from scheduler.redis_models.worker import WorkerModel
from scheduler.settings import (
    SCHEDULER_CONFIG,
    get_queue_names,
    get_queue_configuration,
    QueueConfiguration,
    logger,
)
from scheduler.types import ConnectionErrorTypes, BrokerMetaData, Broker
from .queue_logic import Queue


_BAD_QUEUE_CONFIGURATION = set()


def _get_connection(config: QueueConfiguration, use_strict_broker=False):
    """Returns a Broker connection to use based on parameters in SCHEDULER_QUEUES"""
    if SCHEDULER_CONFIG.BROKER == Broker.FAKEREDIS:
        import fakeredis

        broker_cls = fakeredis.FakeRedis if not use_strict_broker else fakeredis.FakeStrictRedis
    else:
        broker_cls = BrokerMetaData[(SCHEDULER_CONFIG.BROKER, use_strict_broker)].connection_type
    if config.URL:
        return broker_cls.from_url(config.URL, db=config.DB, **(config.CONNECTION_KWARGS or {}))
    if config.UNIX_SOCKET_PATH:
        return broker_cls(unix_socket_path=config.UNIX_SOCKET_PATH, db=config.DB)

    if config.SENTINELS:
        connection_kwargs = {
            "db": config.DB,
            "password": config.PASSWORD,
            "username": config.USERNAME,
        }
        connection_kwargs.update(config.CONNECTION_KWARGS or {})
        sentinel_kwargs = config.SENTINEL_KWARGS or {}
        SentinelClass = BrokerMetaData[(SCHEDULER_CONFIG.BROKER, use_strict_broker)].sentinel_type
        sentinel = SentinelClass(config.SENTINELS, sentinel_kwargs=sentinel_kwargs, **connection_kwargs)
        return sentinel.master_for(
            service_name=config.MASTER_NAME,
            redis_class=broker_cls,
        )

    return broker_cls(
        host=config.HOST,
        port=config.PORT,
        db=config.DB,
        username=config.USERNAME,
        password=config.PASSWORD,
        **(config.CONNECTION_KWARGS or {}),
    )


def get_queue(name="default") -> Queue:
    """Returns an DjangoQueue using parameters defined in `SCHEDULER_QUEUES`"""
    queue_settings = get_queue_configuration(name)
    is_async = queue_settings.ASYNC
    connection = _get_connection(queue_settings)
    return Queue(name=name, connection=connection, is_async=is_async)


def get_all_workers() -> Set[WorkerModel]:
    queue_names = get_queue_names()

    workers_set: Set[WorkerModel] = set()
    for queue_name in queue_names:
        if queue_name in _BAD_QUEUE_CONFIGURATION:
            continue
        connection = _get_connection(get_queue_configuration(queue_name))
        try:
            curr_workers: Set[WorkerModel] = set(WorkerModel.all(connection=connection))
            workers_set.update(curr_workers)
        except ConnectionErrorTypes as e:
            logger.error(f"Could not connect for queue {queue_name}: {e}")
            _BAD_QUEUE_CONFIGURATION.add(queue_name)
    return workers_set



from django.urls import reverse

from scheduler.worker import create_worker
from scheduler.tests import test_settings  # noqa
from scheduler.tests.test_views.base import BaseTestCase


class TestViewWorkers(BaseTestCase):
    def test_workers_home(self):
        res = self.client.get(reverse("workers_home"))
        prev_workers = res.context["workers"]
        worker1 = create_worker("django_tasks_scheduler_test")
        worker1.worker_start()
        worker2 = create_worker("test3")
        worker2.worker_start()

        res = self.client.get(reverse("workers_home"))
        self.assertEqual(res.context["workers"], prev_workers + [worker1._model, worker2._model])



import time

from django.test import TestCase

from scheduler import settings
from scheduler.helpers.queues import get_queue
from . import test_settings  # noqa
from ..decorators import JOB_METHODS_LIST, job
from ..redis_models import JobStatus
from ..redis_models.job import JobModel
from ..worker import create_worker


@job()
def test_job():
    time.sleep(1)
    return 1 + 1


@job("django_tasks_scheduler_test")
def test_job_diff_queue():
    time.sleep(1)
    return 1 + 1


@job(timeout=1)
def test_job_timeout():
    time.sleep(1)
    return 1 + 1


@job(result_ttl=1)
def test_job_result_ttl():
    return 1 + 1


class MyClass:
    def run(self):
        print("Hello")

    def __eq__(self, other):
        if not isinstance(other, MyClass):
            return False
        return True


@job()
def long_running_func(x):
    x.run()


class JobDecoratorTest(TestCase):
    def setUp(self) -> None:
        get_queue("default").connection.flushall()

    def test_all_job_methods_registered(self):
        self.assertEqual(5, len(JOB_METHODS_LIST))

    def test_job_decorator_no_params(self):
        test_job.delay()
        self._assert_job_with_func_and_props(
            "default",
            test_job,
            settings.SCHEDULER_CONFIG.DEFAULT_SUCCESS_TTL,
            settings.SCHEDULER_CONFIG.DEFAULT_JOB_TIMEOUT,
        )

    def test_job_decorator_timeout(self):
        test_job_timeout.delay()
        self._assert_job_with_func_and_props(
            "default",
            test_job_timeout,
            settings.SCHEDULER_CONFIG.DEFAULT_SUCCESS_TTL,
            1,
        )

    def test_job_decorator_result_ttl(self):
        test_job_result_ttl.delay()
        self._assert_job_with_func_and_props(
            "default",
            test_job_result_ttl,
            1,
            settings.SCHEDULER_CONFIG.DEFAULT_JOB_TIMEOUT,
        )

    def test_job_decorator_different_queue(self):
        test_job_diff_queue.delay()
        self._assert_job_with_func_and_props(
            "django_tasks_scheduler_test",
            test_job_diff_queue,
            settings.SCHEDULER_CONFIG.DEFAULT_SUCCESS_TTL,
            settings.SCHEDULER_CONFIG.DEFAULT_JOB_TIMEOUT,
        )

    def _assert_job_with_func_and_props(self, queue_name, expected_func, expected_result_ttl, expected_timeout):
        queue = get_queue(queue_name)
        jobs = JobModel.get_many(queue.queued_job_registry.all(), queue.connection)
        self.assertEqual(1, len(jobs))

        j = jobs[0]
        self.assertEqual(j.func, expected_func)
        self.assertEqual(j.success_ttl, expected_result_ttl)
        self.assertEqual(j.timeout, expected_timeout)

    def test_job_decorator_bad_queue(self):
        with self.assertRaises(settings.QueueNotFoundError):

            @job("bad-queue")
            def test_job_bad_queue():
                time.sleep(1)
                return 1 + 1

    def test_job_decorator_delay_with_param(self):
        queue_name = "default"
        long_running_func.delay(MyClass())

        worker = create_worker(queue_name, burst=True)
        worker.work()

        jobs_list = worker.queues[0].get_all_jobs()
        self.assertEqual(1, len(jobs_list))
        job = jobs_list[0]
        self.assertEqual(job.func, long_running_func)
        self.assertEqual(job.kwargs, {})
        self.assertEqual(job.status, JobStatus.FINISHED)
        self.assertEqual(job.args, (MyClass(),))



import dataclasses
from datetime import datetime
from enum import Enum
from typing import List, Optional, ClassVar, Any, Generator

from scheduler.helpers.utils import utcnow
from scheduler.redis_models.base import HashModel, MAX_KEYS
from scheduler.settings import logger
from scheduler.types import ConnectionType, Self

DEFAULT_WORKER_TTL = 420


class WorkerStatus(str, Enum):
    CREATED = "created"
    STARTING = "starting"
    STARTED = "started"
    SUSPENDED = "suspended"
    BUSY = "busy"
    IDLE = "idle"


@dataclasses.dataclass(slots=True, kw_only=True)
class WorkerModel(HashModel):
    name: str
    queue_names: List[str]
    pid: int
    hostname: str
    ip_address: str
    version: str
    python_version: str
    state: WorkerStatus
    job_execution_process_pid: int = 0
    successful_job_count: int = 0
    failed_job_count: int = 0
    completed_jobs: int = 0
    birth: Optional[datetime] = None
    last_heartbeat: Optional[datetime] = None
    is_suspended: bool = False
    current_job_name: Optional[str] = None
    stopped_job_name: Optional[str] = None
    total_working_time_ms: float = 0.0
    current_job_working_time: float = 0
    last_cleaned_at: Optional[datetime] = None
    shutdown_requested_date: Optional[datetime] = None
    has_scheduler: bool = False
    death: Optional[datetime] = None

    _list_key: ClassVar[str] = ":workers:ALL:"
    _children_key_template: ClassVar[str] = ":queue-workers:{}:"
    _element_key_template: ClassVar[str] = ":workers:{}"

    def save(self, connection: ConnectionType) -> None:
        pipeline = connection.pipeline()
        super(WorkerModel, self).save(pipeline)
        for queue_name in self.queue_names:
            pipeline.sadd(self._children_key_template.format(queue_name), self.name)
        pipeline.expire(self._key, DEFAULT_WORKER_TTL + 60)
        pipeline.execute()

    def delete(self, connection: ConnectionType) -> None:
        logger.debug(f"Deleting worker {self.name}")
        pipeline = connection.pipeline()
        now = utcnow()
        self.death = now
        pipeline.hset(self._key, "death", now.isoformat())
        pipeline.expire(self._key, 60)
        pipeline.srem(self._list_key, self.name)
        for queue_name in self.queue_names:
            pipeline.srem(self._children_key_template.format(queue_name), self.name)
        pipeline.execute()

    def __eq__(self, other: Self) -> bool:
        if not isinstance(other, self.__class__):
            raise TypeError("Cannot compare workers to other types (of workers)")
        return self._key == other._key

    def __hash__(self):
        """The hash does not take the database/connection into account"""
        return hash((self._key, ",".join(self.queue_names)))

    def set_current_job_working_time(self, job_execution_time: int, connection: ConnectionType) -> None:
        self.set_field("current_job_working_time", job_execution_time, connection=connection)

    def heartbeat(self, connection: ConnectionType, timeout: Optional[int] = None) -> None:
        timeout = timeout or DEFAULT_WORKER_TTL + 60
        connection.expire(self._key, timeout)
        now = utcnow()
        self.set_field("last_heartbeat", now, connection=connection)
        logger.debug(f"Next heartbeat for worker {self._key} should arrive in {timeout} seconds.")

    @classmethod
    def cleanup(cls, connection: ConnectionType, queue_name: Optional[str] = None):
        worker_names = cls.all_names(connection, queue_name)
        worker_keys = [cls.key_for(worker_name) for worker_name in worker_names]
        with connection.pipeline() as pipeline:
            for worker_key in worker_keys:
                pipeline.exists(worker_key)
            worker_exist = pipeline.execute()
            invalid_workers = list()
            for i, worker_name in enumerate(worker_names):
                if not worker_exist[i]:
                    invalid_workers.append(worker_name)
            if len(invalid_workers) == 0:
                return
            for invalid_subset in _split_list(invalid_workers, MAX_KEYS):
                pipeline.srem(cls._list_key, *invalid_subset)
                if queue_name:
                    pipeline.srem(cls._children_key_template.format(queue_name), *invalid_subset)
                pipeline.execute()


def _split_list(a_list: List[str], segment_size: int) -> Generator[list[str], Any, None]:
    """Splits a list into multiple smaller lists having size `segment_size`

    :param a_list: The list to split
    :param segment_size: The segment size to split into
    :returns: The list split into smaller lists
    """
    for i in range(0, len(a_list), segment_size):
        yield a_list[i : i + segment_size]



from typing import List

from django.contrib import admin, messages
from django.contrib.contenttypes.admin import GenericStackedInline
from django.db.models import QuerySet
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from scheduler.helpers.queues import get_queue
from scheduler.models import TaskArg, TaskKwarg, Task, TaskType, get_next_cron_time
from scheduler.redis_models import JobModel
from scheduler.settings import SCHEDULER_CONFIG, logger
from scheduler.types import ConnectionErrorTypes


def job_execution_of(job: JobModel, task: Task) -> bool:
    return job.scheduled_task_id == task.id and job.task_type == task.task_type


def get_job_executions_for_task(queue_name: str, scheduled_task: Task) -> List[JobModel]:
    queue = get_queue(queue_name)
    job_list: List[JobModel] = JobModel.get_many(queue.get_all_job_names(), connection=queue.connection)
    res = sorted(
        list(filter(lambda j: job_execution_of(j, scheduled_task), job_list)), key=lambda j: j.created_at, reverse=True
    )
    return res


class JobArgInline(GenericStackedInline):
    model = TaskArg
    extra = 0
    fieldsets = ((None, dict(fields=("arg_type", "val"))),)


class JobKwargInline(GenericStackedInline):
    model = TaskKwarg
    extra = 0
    fieldsets = ((None, dict(fields=("key", ("arg_type", "val")))),)


def get_message_bit(rows_updated: int) -> str:
    message_bit = "1 task was" if rows_updated == 1 else f"{rows_updated} tasks were"
    return message_bit


@admin.register(Task)
class TaskAdmin(admin.ModelAdmin):
    """TaskAdmin admin view for all task models."""

    class Media:
        js = (
            "admin/js/jquery.init.js",
            "admin/js/select-fields.js",
        )

    save_on_top = True
    change_form_template = "admin/scheduler/change_form.html"
    actions = [
        "disable_selected",
        "enable_selected",
        "enqueue_job_now",
    ]
    inlines = [
        JobArgInline,
        JobKwargInline,
    ]
    list_filter = ("enabled", "task_type", "queue")
    list_display = (
        "enabled",
        "name",
        "function_string",
        "is_scheduled",
        "queue",
        "task_schedule",
        "next_run",
        "successful_runs",
        "last_successful_run",
        "failed_runs",
        "last_failed_run",
    )
    list_display_links = ("name",)
    readonly_fields = (
        "job_name",
        "successful_runs",
        "last_successful_run",
        "failed_runs",
        "last_failed_run",
    )
    # radio_fields = {"task_type": admin.HORIZONTAL}
    fieldsets = (
        (
            None,
            dict(
                fields=(
                    "name",
                    "callable",
                    ("enabled", "timeout", "result_ttl"),
                    "task_type",
                )
            ),
        ),
        (
            None,
            dict(fields=("scheduled_time",), classes=("tasktype-OnceTaskType",)),
        ),
        (
            None,
            dict(fields=("cron_string",), classes=("tasktype-CronTaskType",)),
        ),
        (
            None,
            dict(
                fields=(
                    (
                        "interval",
                        "interval_unit",
                    ),
                    "repeat",
                ),
                classes=("tasktype-RepeatableTaskType",),
            ),
        ),
        (_("Queue settings"), dict(fields=(("queue", "at_front"), "job_name"))),
        (
            _("Previous runs info"),
            dict(fields=(("successful_runs", "last_successful_run"), ("failed_runs", "last_failed_run"))),
        ),
    )

    @admin.display(description="Schedule")
    def task_schedule(self, o: Task) -> str:
        if o.task_type == TaskType.ONCE.value:
            return f"Run once: {o.scheduled_time:%Y-%m-%d %H:%M:%S}"
        elif o.task_type == TaskType.CRON.value:
            return f"Cron: {o.cron_string}"
        else:  # if o.task_type == TaskType.REPEATABLE.value:
            if o.interval is None or o.interval_unit is None:
                return ""
            return f"Repeatable: {o.interval} {o.get_interval_unit_display()}"

    @admin.display(description="Next run")
    def next_run(self, o: Task) -> str:
        return get_next_cron_time(o.cron_string)

    def change_view(self, request: HttpRequest, object_id, form_url="", extra_context=None):
        extra = extra_context or {}
        obj = self.get_object(request, object_id)
        try:
            execution_list = get_job_executions_for_task(obj.queue, obj)
        except ConnectionErrorTypes as e:
            logger.warn(f"Could not get job executions: {e}")
            execution_list = list()
        paginator = self.get_paginator(request, execution_list, SCHEDULER_CONFIG.EXECUTIONS_IN_PAGE)
        page_number = request.GET.get("p", 1)
        page_obj = paginator.get_page(page_number)
        page_range = paginator.get_elided_page_range(page_obj.number)

        extra.update(
            {
                "pagination_required": paginator.count > SCHEDULER_CONFIG.EXECUTIONS_IN_PAGE,
                "executions": page_obj,
                "page_range": page_range,
                "page_var": "p",
            }
        )

        return super(TaskAdmin, self).change_view(request, object_id, form_url, extra_context=extra)

    def delete_queryset(self, request: HttpRequest, queryset: QuerySet) -> None:
        for job in queryset:
            job.unschedule()
        super(TaskAdmin, self).delete_queryset(request, queryset)

    def delete_model(self, request: HttpRequest, obj: Task) -> None:
        obj.unschedule()
        super(TaskAdmin, self).delete_model(request, obj)

    @admin.action(description=_("Disable selected %(verbose_name_plural)s"), permissions=("change",))
    def disable_selected(self, request: HttpRequest, queryset: QuerySet) -> None:
        rows_updated = 0
        for obj in queryset.filter(enabled=True).iterator():
            obj.enabled = False
            obj.unschedule()
            rows_updated += 1

        level = messages.WARNING if not rows_updated else messages.INFO
        self.message_user(
            request, f"{get_message_bit(rows_updated)} successfully disabled and unscheduled.", level=level
        )

    @admin.action(description=_("Enable selected %(verbose_name_plural)s"), permissions=("change",))
    def enable_selected(self, request: HttpRequest, queryset: QuerySet) -> None:
        rows_updated = 0
        for obj in queryset.filter(enabled=False).iterator():
            obj.enabled = True
            obj.save()
            rows_updated += 1

        level = messages.WARNING if not rows_updated else messages.INFO
        self.message_user(request, f"{get_message_bit(rows_updated)} successfully enabled and scheduled.", level=level)

    @admin.action(description="Enqueue now", permissions=("change",))
    def enqueue_job_now(self, request: HttpRequest, queryset: QuerySet) -> None:
        task_names = []
        for task in queryset:
            task.enqueue_to_run()
            task_names.append(task.name)
        self.message_user(
            request,
            f"The following jobs have been enqueued: {', '.join(task_names)}",
        )



# This is a helper module to obfuscate types used by different broker implementations.
from collections import namedtuple
from typing import Any, Callable, TypeVar, Union
from typing import Dict, Tuple

import redis

try:
    import valkey
except ImportError:
    valkey = redis
    valkey.Valkey = redis.Redis
    valkey.StrictValkey = redis.StrictRedis

from .settings_types import Broker

ConnectionErrorTypes = (redis.ConnectionError, valkey.ConnectionError)
ResponseErrorTypes = (redis.ResponseError, valkey.ResponseError)
TimeoutErrorTypes = (redis.TimeoutError, valkey.TimeoutError)
WatchErrorTypes = (redis.WatchError, valkey.WatchError)
ConnectionType = Union[redis.Redis, valkey.Valkey]
PipelineType = Union[redis.client.Pipeline, valkey.client.Pipeline]
SentinelType = Union[redis.sentinel.Sentinel, valkey.sentinel.Sentinel]
FunctionReferenceType = TypeVar("FunctionReferenceType", str, Callable[..., Any])

BrokerMetaDataType = namedtuple("BrokerMetaDataType", ["connection_type", "sentinel_type"])

BrokerMetaData: Dict[Tuple[Broker, bool], BrokerMetaDataType] = {
    # Map of (Broker, Strict flag) => Connection Class, Sentinel Class
    (Broker.REDIS, False): BrokerMetaDataType(redis.Redis, redis.sentinel.Sentinel),
    (Broker.VALKEY, False): BrokerMetaDataType(valkey.Valkey, valkey.sentinel.Sentinel),
    (Broker.REDIS, True): BrokerMetaDataType(redis.StrictRedis, redis.sentinel.Sentinel),
    (Broker.VALKEY, True): BrokerMetaDataType(valkey.StrictValkey, valkey.sentinel.Sentinel),
}

TASK_TYPES = ["OnceTaskType", "RepeatableTaskType", "CronTaskType"]



from typing import Optional, List
from typing import Tuple
from urllib.parse import urlparse

from django.contrib import messages
from django.http import Http404
from django.http import HttpRequest
from django.urls import resolve
from django.utils.http import url_has_allowed_host_and_scheme

from scheduler.helpers.queues import Queue
from scheduler.helpers.queues import get_queue as get_queue_base
from scheduler.redis_models import JobModel
from scheduler.settings import QueueNotFoundError
from scheduler.settings import get_queue_names, logger

_QUEUES_WITH_BAD_CONFIGURATION = set()


def get_queue(queue_name: str) -> Queue:
    try:
        return get_queue_base(queue_name)
    except QueueNotFoundError as e:
        logger.error(e)
        raise Http404(e)


def _find_job(job_name: str) -> Tuple[Optional[Queue], Optional[JobModel]]:
    queue_names = get_queue_names()
    for queue_name in queue_names:
        if queue_name in _QUEUES_WITH_BAD_CONFIGURATION:
            continue
        try:
            queue = get_queue(queue_name)
            job = JobModel.get(job_name, connection=queue.connection)
            if job is not None and job.queue_name == queue_name:
                return queue, job
        except Exception as e:
            _QUEUES_WITH_BAD_CONFIGURATION.add(queue_name)
            logger.debug(f"Queue {queue_name} added to bad configuration - Got exception: {e}")
            pass
    return None, None


def _check_next_url(request: HttpRequest, default_next_url: str) -> str:
    next_url = request.POST.get("next_url", default_next_url)
    next_url = next_url.replace("\\", "")
    if (
        not url_has_allowed_host_and_scheme(next_url, allowed_hosts=None)
        or urlparse(next_url).netloc
        or urlparse(next_url).scheme
    ):
        messages.warning(request, "Bad followup URL")
        next_url = default_next_url
    try:
        resolve(next_url)
    except Exception:
        messages.warning(request, "Bad followup URL")
        next_url = default_next_url
    return next_url


def _enqueue_multiple_jobs(queue: Queue, job_names: List[str], at_front: bool = False) -> int:
    jobs = JobModel.get_many(job_names, connection=queue.connection)
    jobs_requeued = 0
    with queue.connection.pipeline() as pipe:
        for job in jobs:
            if job is None:
                continue
            job.save(connection=pipe)
            queue.enqueue_job(job, connection=pipe, at_front=at_front)
            jobs_requeued += 1
        pipe.execute()
    return jobs_requeued



import inspect
from typing import Union, Callable, Any, Optional

from scheduler.helpers.utils import callable_func
from scheduler.timeouts import JobTimeoutException


class CallbackSetupError(Exception):
    pass


class Callback:
    def __init__(self, func: Union[str, Callable[..., Any]], timeout: Optional[int] = None):
        from scheduler.settings import SCHEDULER_CONFIG

        self.timeout = timeout or SCHEDULER_CONFIG.CALLBACK_TIMEOUT
        if not isinstance(self.timeout, int) or self.timeout < 0:
            raise CallbackSetupError(f"Callback `timeout` must be a positive int, but received {self.timeout}")
        if not isinstance(func, str) and not inspect.isfunction(func) and not inspect.isbuiltin(func):
            raise CallbackSetupError(f"Callback `func` must be a string or function, received {func}")
        if isinstance(func, str):
            try:
                func_str = func
                func = callable_func(func)
            except (TypeError, AttributeError, ModuleNotFoundError, ValueError):
                raise CallbackSetupError(f"Callback `func` is not callable: {func_str}")
        self.func: Callable[..., Any] = func

    @property
    def name(self) -> str:
        return f"{self.func.__module__}.{self.func.__qualname__}"

    def __call__(self, *args, **kwargs):
        from scheduler.settings import SCHEDULER_CONFIG

        with SCHEDULER_CONFIG.DEATH_PENALTY_CLASS(self.timeout, JobTimeoutException):
            return self.func(*args, **kwargs)



from enum import Enum
from html import escape

from django.contrib import admin, messages
from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponse, HttpRequest
from django.http.response import HttpResponseBadRequest
from django.shortcuts import render, redirect
from django.views.decorators.cache import never_cache

from scheduler.helpers.queues import InvalidJobOperation
from scheduler.redis_models import Result
from scheduler.settings import logger
from scheduler.views.helpers import _find_job
from scheduler.worker.commands import send_command, StopJobCommand


class JobDetailAction(str, Enum):
    DELETE = "delete"
    ENQUEUE = "enqueue"
    CANCEL = "cancel"


@never_cache
@staff_member_required
def job_detail(request: HttpRequest, job_name: str) -> HttpResponse:
    queue, job = _find_job(job_name)
    if job is None:
        messages.warning(request, f"Job {escape(job_name)} does not exist, maybe its TTL has passed")
        return redirect("queues_home")
    try:
        job.func_name
        data_is_valid = True
    except Exception:
        data_is_valid = False

    try:
        last_result = Result.fetch_latest(queue.connection, job.name)
    except AttributeError:
        last_result = None

    context_data = {
        **admin.site.each_context(request),
        "job": job,
        "last_result": last_result,
        "results": Result.all(connection=queue.connection, parent=job.name),
        "queue": queue,
        "data_is_valid": data_is_valid,
    }
    return render(request, "admin/scheduler/job_detail.html", context_data)


@never_cache
@staff_member_required
def job_action(request: HttpRequest, job_name: str, action: str) -> HttpResponse:
    queue, job = _find_job(job_name)
    if job is None:
        messages.warning(request, f"Job {escape(job_name)} does not exist, maybe its TTL has passed")
        return redirect("queues_home")
    if action not in [item.value for item in JobDetailAction]:
        return HttpResponseBadRequest(f"Action {escape(action)} is not supported")

    if request.method != "POST":
        context_data = {
            **admin.site.each_context(request),
            "job": job,
            "queue": queue,
            "action": action,
        }
        return render(request, "admin/scheduler/single_job_action.html", context_data)

    try:
        if action == JobDetailAction.DELETE:
            queue.delete_job(job.name)
            messages.info(request, f"You have successfully deleted {job.name}")
            return redirect("queue_registry_jobs", queue.name, "queued")
        elif action == JobDetailAction.ENQUEUE:
            queue.delete_job(job.name, expire_job_model=False)
            queue.enqueue_job(job)
            messages.info(request, f"You have successfully enqueued {job.name}")
            return redirect("job_details", job_name)
        elif action == JobDetailAction.CANCEL:
            send_command(
                connection=queue.connection, command=StopJobCommand(job_name=job.name, worker_name=job.worker_name)
            )
            queue.cancel_job(job.name)
            messages.info(request, f"You have successfully cancelled {job.name}")
            return redirect("job_details", job_name)
    except InvalidJobOperation as e:
        logger.warning(f"Could not perform action: {e}")
        messages.warning(request, f"Could not perform action: {e}")
    return redirect("job_details", job_name)
